from mindsdb.integrations.libs.const import HANDLER_TYPE
from mindsdb.utilities import log

logger = log.getLogger(__name__)

from .__about__ import __version__ as version, __description__ as description
try:
    from .cohere_handler import CohereHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = 'Cohere'
name = 'cohere'
type = HANDLER_TYPE.ML

__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'description', 'import_error'
]



import pandas as pd
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb_sql.parser import ast
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class SearchLocationTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        """Select data from the search_location table and return it as a pandas DataFrame.

        Args:
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        """

        self.handler.connect()

        conditions = extract_comparison_conditions(query.where)

        allowed_keys = set(
            [
                "searchQuery",
                "category",
                "phone",
                "address",
                "latLong",
                "radius",
                "radiusUnit",
                "language",
            ]
        )

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")
            elif op == "=" and arg1 in allowed_keys:
                params[arg1] = arg2
            elif op != "=":
                raise NotImplementedError(f"Unknown op: {op}")
            else:
                filters.append([op, arg1, arg2])

        if query.limit is not None:
            params["max_results"] = query.limit.value

        if "searchQuery" not in params and "latLong" not in params:
            # search not works without searchQuery, use 'London'
            params["searchQuery"] = "London"

        result = self.handler.call_tripadvisor_searchlocation_api(params=params)

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        """Get the list of column names for the search_location table.

        Returns:
            list: A list of column names for the search_location table.
        """
        return [
            "location_id",
            "name",
            "distance",
            "rating",
            "bearing",
            "street1",
            "street2",
            "city",
            "state",
            "country",
            "postalcode",
            "address_string",
            "phone",
            "latitude",
            "longitude",
        ]


class LocationDetailsTable(APITable):
    result_json = []

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Select data from the location_details table and return it as a pandas DataFrame.

        Args:
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        """

        self.handler.connect()

        conditions = extract_comparison_conditions(query.where)

        allowed_keys = set(["locationId", "currency", "language"])

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")
            elif op == "=" and arg1 in allowed_keys:
                params[arg1] = arg2
            elif op != "=":
                raise NotImplementedError(f"Unknown op: {op}")
            else:
                filters.append([op, arg1, arg2])

        if query.limit is not None:
            params["max_results"] = query.limit.value

        if "locationId" not in params:
            # search not works without searchQuery, use 'London'
            params["locationId"] = "23322232"

        result = self.handler.call_tripadvisor_location_details_api(params=params)

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        """Get the list of column names for the location_details table.

        Returns:
            list: A list of column names for the location_details table.
        """
        return [
            "location_id",
            "distance",
            "name",
            "description",
            "web_url",
            "street1",
            "street2",
            "city",
            "state",
            "country",
            "postalcode",
            "address_string",
            "latitude",
            "longitude",
            "timezone",
            "email",
            "phone",
            "website",
            "write_review",
            "ranking_data",
            "rating",
            "rating_image_url",
            "num_reviews",
            "photo_count",
            "see_all_photos",
            "price_level",
            "brand",
            "parent_brand",
            "ancestors",
            "periods",
            "weekday",
            "features",
            "cuisines",
            "amenities",
            "trip_types",
            "styles",
            "awards",
            "neighborhood_info",
            "parent_brand",
            "brand",
            "groups",
        ]


class ReviewsTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        """Select data from the reviews table and return it as a pandas DataFrame.

        Args:
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        """

        self.handler.connect()

        conditions = extract_comparison_conditions(query.where)

        allowed_keys = set(["locationId", "language"])

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")
            elif op == "=" and arg1 in allowed_keys:
                params[arg1] = arg2
            elif op != "=":
                raise NotImplementedError(f"Unknown op: {op}")
            else:
                filters.append([op, arg1, arg2])

        if query.limit is not None:
            params["max_results"] = query.limit.value

        if "locationId" not in params:
            # search not works without searchQuery, use 'London'
            params["locationId"] = "23322232"

        result = self.handler.call_tripadvisor_reviews_api(params=params)

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        """Get the list of column names for the reviews table.

        Returns:
            list: A list of column names for the reviews table.
        """
        return [
            "id",
            "lang",
            "location_id",
            "published_date",
            "rating",
            "helpful_votes",
            "rating_image_url",
            "url",
            "trip_type",
            "travel_date",
            "text_review",
            "title",
            "owner_response",
            "is_machine_translated",
            "user",
            "subratings",
        ]


class PhotosTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        """Select data from the photos table and return it as a pandas DataFrame.

        Args:
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        """

        conditions = extract_comparison_conditions(query.where)

        allowed_keys = set(["locationId", "language"])

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")
            elif op == "=" and arg1 in allowed_keys:
                params[arg1] = arg2
            elif op != "=":
                raise NotImplementedError(f"Unknown op: {op}")
            else:
                filters.append([op, arg1, arg2])

        if query.limit is not None:
            params["max_results"] = query.limit.value

        if "locationId" not in params:
            params["locationId"] = "23322232"

        result = self.handler.call_tripadvisor_photos_api(params=params)

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        """Get the list of column names for the photos table.

        Returns:
            list: A list of column names for the photos table.
        """
        return [
            "id",
            "is_blessed",
            "album",
            "caption",
            "published_date",
            "images",
            "source",
            "user",
        ]


class NearbyLocationTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        """Select data from the nearby_location table and return it as a pandas DataFrame.

        Args:
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        """

        conditions = extract_comparison_conditions(query.where)

        allowed_keys = set(["latLong", "language", "category", "phone", "address", "radius", "radiusUnit"])

        params = {}
        filters = []
        for op, arg1, arg2 in conditions:
            if op == "or":
                raise NotImplementedError("OR is not supported")
            elif op == "=" and arg1 in allowed_keys:
                params[arg1] = arg2
            elif op != "=":
                raise NotImplementedError(f"Unknown op: {op}")
            else:
                filters.append([op, arg1, arg2])

        if query.limit is not None:
            params["max_results"] = query.limit.value

        if "latLong" not in params:
            params["latLong"] = "40.780825, -73.972781"

        result = self.handler.call_tripadvisor_nearby_location_api(params=params)

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = []
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError

        if len(columns) == 0:
            columns = self.get_columns()

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            # add absent columns
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            # filter by columns
            result = result[columns]
        return result

    def get_columns(self):
        """Get the list of column names for the nearby_location table.

        Returns:
            list: A list of column names for the nearby_location table.
        """
        return [
            "location_id",
            "name",
            "distance",
            "rating",
            "bearing",
            "address_obj",
        ]



import unittest
from mindsdb.integrations.handlers.google_content_shopping_handler.google_content_shopping_handler import \
    GoogleContentShoppingHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class GoogleSearchConsoleHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "merchant_id": "1234567890",
                "credentials": "/path/to/credentials.json"
            },
            "file_storage": "/path/to/file_storage"
        }
        cls.handler = GoogleContentShoppingHandler('test_google_content_shopping_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_2_select_accounts_query(self):
        query = "SELECT * FROM accounts LIMIT 10"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_3_select_orders_query(self):
        query = "SELECT kind FROM orders WHERE orderId > 100"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_4_select_products_query(self):
        query = "SELECT price FROM products WHERE brand = 'Google' LIMIT 100"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_5_delete_accounts_query(self):
        query = "DELETE FROM accounts WHERE accountId = '1234567890'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_6_delete_orders_query(self):
        query = "DELETE FROM orders WHERE orderId < '1234567890'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_7_delete_products_query(self):
        query = "DELETE FROM products WHERE brand = 'Google'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_8_update_products_query(self):
        query = "UPDATE products SET price = 100 WHERE productId > 120 AND updateMask = 'price'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE


if __name__ == '__main__':
    unittest.main()



import mindsdb.microservices_grpc.db.common_pb2 as _common_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class BinaryQueryContext(_message.Message):
    __slots__ = ["context", "query"]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    context: HandlerContext
    query: bytes
    def __init__(self, context: _Optional[_Union[HandlerContext, _Mapping]] = ..., query: _Optional[bytes] = ...) -> None: ...

class ColumnsContext(_message.Message):
    __slots__ = ["context", "table"]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    TABLE_FIELD_NUMBER: _ClassVar[int]
    context: HandlerContext
    table: str
    def __init__(self, context: _Optional[_Union[HandlerContext, _Mapping]] = ..., table: _Optional[str] = ...) -> None: ...

class HandlerContext(_message.Message):
    __slots__ = ["context", "handler_params", "handler_type"]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    HANDLER_PARAMS_FIELD_NUMBER: _ClassVar[int]
    HANDLER_TYPE_FIELD_NUMBER: _ClassVar[int]
    context: str
    handler_params: str
    handler_type: str
    def __init__(self, context: _Optional[str] = ..., handler_type: _Optional[str] = ..., handler_params: _Optional[str] = ...) -> None: ...

class NativeQueryContext(_message.Message):
    __slots__ = ["context", "query"]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    QUERY_FIELD_NUMBER: _ClassVar[int]
    context: HandlerContext
    query: str
    def __init__(self, context: _Optional[_Union[HandlerContext, _Mapping]] = ..., query: _Optional[str] = ...) -> None: ...



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class BinaryExecutionContext(_message.Message):
    __slots__ = ["context", "sql"]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    SQL_FIELD_NUMBER: _ClassVar[int]
    context: ExecutorContext
    sql: bytes
    def __init__(self, context: _Optional[_Union[ExecutorContext, _Mapping]] = ..., sql: _Optional[bytes] = ...) -> None: ...

class DefaultDBContext(_message.Message):
    __slots__ = ["context", "new_db"]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    NEW_DB_FIELD_NUMBER: _ClassVar[int]
    context: ExecutorContext
    new_db: str
    def __init__(self, context: _Optional[_Union[ExecutorContext, _Mapping]] = ..., new_db: _Optional[str] = ...) -> None: ...

class ExecutionContext(_message.Message):
    __slots__ = ["context", "sql"]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    SQL_FIELD_NUMBER: _ClassVar[int]
    context: ExecutorContext
    sql: str
    def __init__(self, context: _Optional[_Union[ExecutorContext, _Mapping]] = ..., sql: _Optional[str] = ...) -> None: ...

class ExecutorContext(_message.Message):
    __slots__ = ["connection_id", "context", "id", "session", "session_id"]
    CONNECTION_ID_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    SESSION_FIELD_NUMBER: _ClassVar[int]
    SESSION_ID_FIELD_NUMBER: _ClassVar[int]
    connection_id: int
    context: str
    id: str
    session: str
    session_id: str
    def __init__(self, id: _Optional[str] = ..., connection_id: _Optional[int] = ..., session_id: _Optional[str] = ..., session: _Optional[str] = ..., context: _Optional[str] = ...) -> None: ...

class ExecutorResponse(_message.Message):
    __slots__ = ["columns", "data", "error_code", "error_message", "is_executed", "params", "server_status", "session", "state_track"]
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    ERROR_CODE_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    IS_EXECUTED_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    SERVER_STATUS_FIELD_NUMBER: _ClassVar[int]
    SESSION_FIELD_NUMBER: _ClassVar[int]
    STATE_TRACK_FIELD_NUMBER: _ClassVar[int]
    columns: bytes
    data: bytes
    error_code: int
    error_message: str
    is_executed: bool
    params: bytes
    server_status: str
    session: str
    state_track: str
    def __init__(self, columns: _Optional[bytes] = ..., params: _Optional[bytes] = ..., data: _Optional[bytes] = ..., state_track: _Optional[str] = ..., server_status: _Optional[str] = ..., is_executed: bool = ..., session: _Optional[str] = ..., error_code: _Optional[int] = ..., error_message: _Optional[str] = ...) -> None: ...

class ExecutorStatusResponse(_message.Message):
    __slots__ = ["error_message", "success"]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    error_message: str
    success: bool
    def __init__(self, success: bool = ..., error_message: _Optional[str] = ...) -> None: ...

class StatementExecuteContext(_message.Message):
    __slots__ = ["context", "params"]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    PARAMS_FIELD_NUMBER: _ClassVar[int]
    context: ExecutorContext
    params: str
    def __init__(self, context: _Optional[_Union[ExecutorContext, _Mapping]] = ..., params: _Optional[str] = ...) -> None: ...



import json
import pandas as pd
from collections import OrderedDict

from google_auth_oauthlib.flow import InstalledAppFlow
from pandas import DataFrame
from google.oauth2 import service_account
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from .google_search_tables import SearchAnalyticsTable, SiteMapsTable
from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.utilities import log

logger = log.getLogger(__name__)

class GoogleSearchConsoleHandler(APIHandler):
    """
        A class for handling connections and interactions with the Google Search Console API.
    """
    name = 'google_search'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the Google Search Console API handler.
        Args:
            name (str): name of the handler
            kwargs (dict): additional arguments
        """
        super().__init__(name)
        self.token = None
        self.service = None
        self.connection_data = kwargs.get('connection_data', {})
        self.fs_storage = kwargs['file_storage']
        self.credentials_file = self.connection_data.get('credentials', None)
        self.credentials = None
        self.scopes = ['https://www.googleapis.com/auth/webmasters.readonly',
                       'https://www.googleapis.com/auth/webmasters']
        self.is_connected = False
        analytics = SearchAnalyticsTable(self)
        self.analytics = analytics
        self._register_table('Analytics', analytics)
        sitemaps = SiteMapsTable(self)
        self.sitemaps = sitemaps
        self._register_table('Sitemaps', sitemaps)

    def connect(self):
        """
        Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected is True:
            return self.service
        if self.credentials_file:
            try:
                json_str_bytes = self.fs_storage.file_get('token_search.json')
                json_str = json_str_bytes.decode()
                self.credentials = Credentials.from_authorized_user_info(info=json.loads(json_str), scopes=self.scopes)
            except Exception:
                self.credentials = None

            if not self.credentials or not self.credentials.valid:
                if self.credentials and self.credentials.expired and self.credentials.refresh_token:
                    self.credentials.refresh(Request())
                else:
                    self.credentials = Credentials.from_authorized_user_file(
                        self.credentials_file, scopes=self.scopes)
            # Save the credentials for the next run
            json_str = self.credentials.to_json()
            self.fs_storage.file_set('token_search.json', json_str.encode())

            self.service = build('webmasters', 'v3', credentials=self.credentials)
        return self.service

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        response = StatusResponse(False)

        try:
            service = self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Google Search Console API: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str = None) -> Response:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, api's json etc)
        Returns:
            HandlerResponse
        """
        method_name, params = FuncParser().from_string(query)

        df = self.call_application_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def get_traffic_data(self, params: dict = None) -> DataFrame:
        """
        Get traffic data from Google Search Console API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        accepted_params = ['start_date', 'end_date', 'dimensions', 'row_limit', 'aggregation_type']
        search_analytics_query_request = {
            key: value for key, value in params.items() if key in accepted_params and value is not None
        }
        response = service.searchanalytics(). \
            query(siteUrl=params['siteUrl'], body=search_analytics_query_request). \
            execute()
        df = pd.DataFrame(response['rows'], columns=self.analytics.get_columns())
        return df

    def get_sitemaps(self, params: dict = None) -> DataFrame:
        """
        Get sitemaps data from Google Search Console API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        if params['sitemapIndex']:
            response = service.sitemaps().list(siteUrl=params['siteUrl'], sitemapIndex=params['sitemapIndex']).execute()
        else:
            response = service.sitemaps().list(siteUrl=params['siteUrl']).execute()
        df = pd.DataFrame(response['sitemap'], columns=self.sitemaps.get_columns())

        # Get as many sitemaps as indicated by the row_limit parameter
        if params['row_limit']:
            if params['row_limit'] > len(df):
                row_limit = len(df)
            else:
                row_limit = params['row_limit']

            df = df[:row_limit]

        return df

    def submit_sitemap(self, params: dict = None) -> DataFrame:
        """
        Submit sitemap to Google Search Console API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        response = service.sitemaps().submit(siteUrl=params['siteUrl'], feedpath=params['feedpath']).execute()
        df = pd.DataFrame(response, columns=self.sitemaps.get_columns())
        return df

    def delete_sitemap(self, params: dict = None) -> DataFrame:
        """
        Delete sitemap from Google Search Console API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        response = service.sitemaps().delete(siteUrl=params['siteUrl'], feedpath=params['feedpath']).execute()
        df = pd.DataFrame(response, columns=self.sitemaps.get_columns())
        return df

    def call_application_api(self, method_name: str = None, params: dict = None) -> DataFrame:
        """
        Call Google Search Console API and map the data to pandas DataFrame
        Args:
            method_name (str): method name
            params (dict): query parameters
        Returns:
            DataFrame
        """
        if method_name == 'get_traffic_data':
            return self.get_traffic_data(params)
        elif method_name == 'get_sitemaps':
            return self.get_sitemaps(params)
        elif method_name == 'submit_sitemap':
            return self.submit_sitemap(params)
        elif method_name == 'delete_sitemap':
            return self.delete_sitemap(params)
        else:
            raise NotImplementedError(f'Unknown method {method_name}')


connection_args = OrderedDict(
    credentials={
        'type': ARG_TYPE.PATH,
        'description': 'The path to the credentials file. If not specified, the default credentials are used.'
    }
)

connection_args_example = OrderedDict(
    credentials='/path/to/credentials.json'
)



from collections import OrderedDict
from typing import Optional
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb_sql import parse_sql
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
import pandas as pd
import jaydebeapi as jdbcconnector

logger = log.getLogger(__name__)

class DerbyHandler(DatabaseHandler):


    name= 'derby'


    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """ Initialize the handler
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        
        self.kwargs = kwargs
        self.parser = parse_sql
        self.database = connection_data['database']
        self.connection_config = connection_data
        self.host = connection_data['host']
        self.port = connection_data['port']
        self.schema =  'APP'
        self.connection = None
        self.is_connected = False
          
    
    def connect(self):
        """ Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            Connection Object
        """
        if self.is_connected is True:
            return self.connection

        user = self.connection_config.get('user')
        password = self.connection_config.get('password')
        jdbc_class = self.connection_config.get('jdbcClass')
        jar_location = self.connection_config.get('jdbcJarLocation')

        jdbc_url = "jdbc:derby://" + self.host + ":" + self.port + "/" + self.database + ";"

        if not jdbc_class: 
            jdbc_class = "org.apache.derby.jdbc.ClientDriver"

        if user: 
            self.schema = user

        try:
            if user and password and jar_location: 
                self.connection = jdbcconnector.connect(jclassname=jdbc_class, url=jdbc_url, driver_args=[user, password], jars=jar_location.split(","))
            elif user and password: 
                self.connection = jdbcconnector.connect(jclassname=jdbc_class, url=jdbc_url, driver_args=[user, password])
            elif jar_location: 
                self.connection = jdbcconnector.connect(jclassname=jdbc_class, url=jdbc_url, jars=jar_location.split(","))
            else:
                self.connection = jdbcconnector.connect(jdbc_class, jdbc_url)
        except Exception as e:
            logger.error(f"Error while connecting to {self.database}, {e}")

        return self.connection


    def disconnect(self):
        """ Close any existing connections
        Should switch self.is_connected.
        """
        if self.is_connected is False:
            return
        try:
            self.connection.close()
            self.is_connected=False
        except Exception as e:
            logger.error(f"Error while disconnecting to {self.database}, {e}")

        return 


    def check_connection(self) -> StatusResponse:
        """ Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        responseCode = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            responseCode.success = True
        except Exception as e:
            logger.error(f'Error connecting to database {self.database}, {e}!')
            responseCode.error_message = str(e)
        finally:
            if responseCode.success is True and need_to_close:
                self.disconnect()
            if responseCode.success is False and self.is_connected is True:
                self.is_connected = False

        return responseCode


    def native_query(self, query: str) -> StatusResponse:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)
        Returns:
            HandlerResponse
        """
        need_to_close = self.is_connected is False
        conn = self.connect()
        with conn.cursor() as cur:
            try:
                cur.execute(query)
                if cur.description:
                    result = cur.fetchall() 
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                self.connection.commit()
            except Exception as e:
                logger.error(f'Error running query: {query} on {self.database}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )
                self.connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    
    def query(self, query: ASTNode) -> StatusResponse:
        """Render and execute a SQL query.

        Args:
            query (ASTNode): The SQL query.

        Returns:
            Response: The query result.
        """
        if isinstance(query, ASTNode):
            query_str = query.to_string()
        else:
            query_str = str(query)

        return self.native_query(query_str)


    def get_tables(self) -> StatusResponse:
        """Get a list of all the tables in the database.

        Returns:
            Response: Names of the tables in the database.
        """
        query = f'''
        SELECT st.tablename FROM sys.systables st LEFT OUTER JOIN sys.sysschemas ss ON (st.schemaid = ss.schemaid) WHERE ss.schemaname ='{self.schema}' '''
    
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    
    def get_columns(self, table_name: str) -> StatusResponse:
        """Get details about a table.

        Args:
            table_name (str): Name of the table to retrieve details of.

        Returns:
            Response: Details of the table.
        """

        query = f''' SELECT COLUMNNAME FROM SYS.SYSCOLUMNS INNER JOIN SYS.SYSTABLES ON SYS.SYSCOLUMNS.REFERENCEID=SYS.SYSTABLES.TABLEID WHERE TABLENAME='{table_name}' '''
        return self.native_query(query)
    
    
    connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Apache Derby server/database.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect to Apache Derby.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': """
            The database name to use when connecting with the Apache Derby server.
        """
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Apache Derby server. If specified this is also treated as the schema.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Apache Derby server.'
    },
    jdbcClass={
        'type': ARG_TYPE.STR,
        'description': 'The jdbc class which should be used to establish the connection, the default value is:  org.apache.derby.jdbc.ClientDriver.'
    },
    jdbcJarLocation={
        'type': ARG_TYPE.STR,
        'description': 'The location of the jar files which contain the JDBC class. This need not be specified if the required classes are already added to the CLASSPATH variable.'
    }

)


connection_args_example = OrderedDict(
    host='localhost',
    port='1527',
    user='test',
    password='test',
    database="testdb",
    jdbcClass='org.apache.derby.jdbc.ClientDriver',
    jdbcJarLocation='/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derby.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbyclient.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbynet.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbyoptionaltools.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbyrun.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbyshared.jar,/opt/homebrew/Cellar/derby/10.16.1.1/libexec/lib/derbytools.jar',
)




from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.err_packet import ErrPacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.handshake_packet import HandshakePacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.fast_auth_fail_packet import FastAuthFail
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.password_answer import PasswordAnswer
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.handshake_response_packet import HandshakeResponsePacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.ok_packet import OkPacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.switch_auth_packet import SwitchOutPacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.switch_auth_response_packet import SwitchOutResponse
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.command_packet import CommandPacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.column_count_packet import ColumnCountPacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.column_definition_packet import ColumnDefenitionPacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.resultset_row_package import ResultsetRowPacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.eof_packet import EofPacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.stmt_prepare_header import STMTPrepareHeaderPacket
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packets.binary_resultset_row_package import BinaryResultsetRowPacket



from typing import List
import pandas as pd
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb_sql.parser import ast
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb_sql.parser.ast.select.constant import Constant
import json


class StatusPages(APITable):

    # table name in the database
    name = 'status_pages'

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Select

        Returns:
            pd.DataFrame
        """
        conditions = extract_comparison_conditions(query.where)
        # Get page id from query
        _id = None
        for op, arg1, arg2 in conditions:
            if arg1 == 'id' and op == '=':
                _id = arg2
            else:
                raise NotImplementedError

        # Get column names from query
        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        # 'id' needs to selected when searching with 'id'
        temp_selected_columns = selected_columns
        if _id and 'id' not in selected_columns:
            selected_columns = ['id'] + selected_columns

        # Get limit from query
        limit = query.limit.value if query.limit else 20
        total_results = limit

        page_no = 1  # default page no
        result_df = pd.DataFrame(columns=selected_columns)

        # call instatus api and get the response as pd.DataFrame
        while True:
            df = self.handler.call_instatus_api(endpoint='/v2/pages', params={'page': page_no, 'per_page': 100})
            if len(df) == 0 or limit <= 0:
                break
            else:
                result_df = pd.concat([result_df, df[selected_columns]], ignore_index=True)

            page_no += 1
            limit -= len(df)

        # select columns from pandas data frame df
        if result_df.empty:
            result_df = pd.DataFrame(columns=selected_columns)
        elif _id:
            result_df = result_df[result_df['id'] == _id]

        # delete 'id' column if 'id' not present in temp_selected_columns
        if 'id' not in temp_selected_columns and 'id' in selected_columns:
            result_df = result_df.drop('id', axis=1)

        return result_df.head(n=total_results)

    def insert(self, query: ast.Insert) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Insert

        Returns:
            None
        """
        data = {}
        for column, value in zip(query.columns, query.values[0]):
            if isinstance(value, Constant):
                data[column.name] = value.value
            else:
                data[column.name] = value
        self.handler.call_instatus_api(endpoint='/v1/pages', method='POST', data=json.dumps(data))

    def update(self, query: ast.Update) -> None:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. Usually it should be ast.Update
        Returns:
            None
        """
        conditions = extract_comparison_conditions(query.where)
        # Get page id from query
        _id = None
        for op, arg1, arg2 in conditions:
            if arg1 == 'id' and op == '=':
                _id = arg2
            else:
                raise NotImplementedError

        data = {}
        for key, value in query.update_columns.items():
            if isinstance(value, Constant):
                if key == 'components':
                    data[key] = json.loads(value.value)  # Convert 'components' value to a Python list
                else:
                    data[key] = value.value
        self.handler.call_instatus_api(endpoint=f'/v2/{_id}', method='PUT', data=json.dumps(data))

    def get_columns(self, ignore: List[str] = []) -> List[str]:
        """columns

        Args:
            ignore (List[str], optional): exclusion items. Defaults to [].

        Returns:
            List[str]: available columns with `ignore` items removed from the list.
        """
        return [
            "id",
            "subdomain",
            "name",
            "workspaceId",
            "logoUrl",
            "faviconUrl",
            "websiteUrl",
            "customDomain",
            "publicEmail",
            "twitter",
            "status",
            "subscribeBySms",
            "sendSmsNotifications",
            "language",
            "useLargeHeader",
            "brandColor",
            "okColor",
            "disruptedColor",
            "degradedColor",
            "downColor",
            "noticeColor",
            "unknownColor",
            "googleAnalytics",
            "smsService",
            "htmlInMeta",
            "htmlAboveHeader",
            "htmlBelowHeader",
            "htmlAboveFooter",
            "htmlBelowFooter",
            "htmlBelowSummary",
            "uptimeDaysDisplay",
            "uptimeOutageDisplay",
            "launchDate",
            "cssGlobal",
            "onboarded",
            "createdAt",
            "updatedAt"
        ]



from typing import Dict, List

from mindsdb.interfaces.agents.agents_controller import AgentsController
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.storage import db

from mindsdb.utilities.context import context as ctx

from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
from mindsdb.utilities.config import Config


class ChatBotController:
    '''Handles CRUD operations at the database level for Chatbots'''

    OBJECT_TYPE = 'chatbot'

    def __init__(self, project_controller: ProjectController = None, agents_controller: AgentsController = None):
        if project_controller is None:
            project_controller = ProjectController()
        if agents_controller is None:
            agents_controller = AgentsController()
        self.project_controller = project_controller
        self.agents_controller = agents_controller

    def get_chatbot(self, chatbot_name: str, project_name: str = 'mindsdb') -> db.ChatBots:
        '''
        Gets a chatbot by name.

        Parameters:
            chatbot_name (str): The name of the chatbot
            project_name (str): The name of the containing project

        Returns:
            bot (db.ChatBots): The database chatbot object
        '''

        project = self.project_controller.get(name=project_name)

        query = db.session.query(
            db.ChatBots, db.Tasks
        ).join(
            db.Tasks, db.ChatBots.id == db.Tasks.object_id
        ).filter(
            db.ChatBots.name == chatbot_name,
            db.ChatBots.project_id == project.id,
            db.Tasks.object_type == self.OBJECT_TYPE,
            db.Tasks.company_id == ctx.company_id,
        )

        query_result = query.first()
        if query_result is None:
            return None
        bot, task = query_result

        # Include DB, Agent, and Task information in response.
        session = SessionController()
        database_names = {
            i['id']: i['name']
            for i in session.database_controller.get_list()
        }

        agent = self.agents_controller.get_agent_by_id(bot.agent_id)
        agent_obj = agent.as_dict() if agent is not None else None
        bot_obj = {
            'id': bot.id,
            'name': bot.name,
            'project': project_name,
            'agent': agent_obj,
            'database_id': bot.database_id,  # TODO remove in future
            'database': database_names.get(bot.database_id, '?'),
            'model_name': bot.model_name,
            'params': bot.params,
            'created_at': bot.created_at,
            'is_running': task.active,
            'last_error': task.last_error,
        }
        return bot_obj

    def get_chatbots(self, project_name: str = 'mindsdb') -> List[dict]:
        '''
        Gets all chatbots in a project.

        Parameters:
            project_name (str): The name of the containing project

        Returns:
            all_bots (List[db.ChatBots]): List of database chatbot object
        '''

        project = self.project_controller.get(name=project_name)

        query = db.session.query(
            db.ChatBots, db.Tasks
        ).join(
            db.Tasks, db.ChatBots.id == db.Tasks.object_id
        ).filter(
            db.ChatBots.project_id == project.id,
            db.Tasks.object_type == self.OBJECT_TYPE,
            db.Tasks.company_id == ctx.company_id,
        )

        session = SessionController()
        database_names = {
            i['id']: i['name']
            for i in session.database_controller.get_list()
        }

        bots = []
        for bot, task in query.all():
            agent = self.agents_controller.get_agent_by_id(bot.agent_id)
            agent_obj = agent.as_dict() if agent is not None else None
            bots.append(
                {
                    'id': bot.id,
                    'name': bot.name,
                    'project': project_name,
                    'agent': agent_obj,
                    'database_id': bot.database_id,  # TODO remove in future
                    'database': database_names.get(bot.database_id, '?'),
                    'model_name': bot.model_name,
                    'params': bot.params,
                    'created_at': bot.created_at,
                    'is_running': task.active,
                    'last_error': task.last_error,
                }
            )

        return bots

    def add_chatbot(
            self,
            name: str,
            project_name: str,
            model_name: str = None,
            agent_name: str = None,
            database_id: int = None,
            is_running: bool = True,
            params: Dict[str, str] = {}) -> db.ChatBots:
        '''
        Adds a chatbot to the database.

        Parameters:
            name (str): The name of the new chatbot
            project_name (str): The containing project
            model_name (str): The name of the existing ML model the chatbot will use
            agent_name (str): The name of the existing agent the chatbot will use
            database_id (int): The ID of the existing database the chatbot will use
            is_running (bool): Whether or not to start the chatbot right after creation
            params: (Dict[str, str]): Parameters to use when running the chatbot

        Returns:
            bot (db.ChatBots): The created chatbot
        '''

        config = Config()

        is_cloud = config.get('cloud', False)
        if is_cloud and ctx.user_class == 0:
            raise Exception("You can't create chatbot")

        if project_name is None:
            project_name = 'mindsdb'
        project = self.project_controller.get(name=project_name)

        bot = self.get_chatbot(name, project_name)

        if bot is not None:
            raise Exception(f'Chat bot already exists: {name}')

        # check database
        session_controller = SessionController()
        db_record = session_controller.integration_controller.get_by_id(database_id)
        if db_record is None:
            raise Exception(f"Database doesn't exist: {database_id}")

        if model_name is None and agent_name is None:
            raise ValueError('Need to provide either "model_name" or "agent_name" when creating a chatbot')
        if agent_name is not None:
            agent = self.agents_controller.get_agent(agent_name, project_name)
            if agent is None:
                raise ValueError(f"Agent with name doesn't exist: {agent_name}")
        else:
            # Create a new agent with the given model name.
            agent = self.agents_controller.add_agent(name, project_name, model_name, [])

        bot = db.ChatBots(
            name=name,
            project_id=project.id,
            agent_id=agent.id,
            model_name=agent.model_name,
            database_id=database_id,
            params=params,
        )
        db.session.add(bot)
        db.session.flush()

        task_record = db.Tasks(
            company_id=ctx.company_id,
            user_class=ctx.user_class,

            object_type=self.OBJECT_TYPE,
            object_id=bot.id,
            active=is_running
        )
        db.session.add(task_record)

        db.session.commit()

        return bot

    def update_chatbot(
            self,
            chatbot_name: str,
            project_name: str = 'mindsdb',
            name: str = None,
            model_name: str = None,
            agent_name: str = None,
            database_id: int = None,
            is_running: bool = None,
            params: Dict[str, str] = None):
        '''
        Updates a chatbot in the database, creating it if it doesn't already exist.

        Parameters:
            chatbot_name (str): The name of the new chatbot, or existing chatbot to update
            project_name (str): The containing project
            name (str): The updated name of the chatbot
            model_name (str): The name of the existing ML model the chatbot will use
            agent_name (str): The name of the existing agent the chatbot will use
            database_id (int): The ID of the existing database the chatbot will use
            is_running (bool): Whether or not the chatbot will run after update/creation
            params: (Dict[str, str]): Parameters to use when running the chatbot

        Returns:
            bot (db.ChatBots): The created or updated chatbot
        '''

        existing_chatbot = self.get_chatbot(chatbot_name, project_name=project_name)
        if existing_chatbot is None:
            raise Exception(f'Chat bot not found: {chatbot_name}')

        existing_chatbot_rec = db.ChatBots.query.get(existing_chatbot['id'])

        if name is not None and name != chatbot_name:
            # check new name
            bot2 = self.get_chatbot(name, project_name=project_name)
            if bot2 is not None:
                raise Exception(f'Chat already exists: {name}')

            existing_chatbot_rec.name = name

        if agent_name is not None:
            agent = self.agents_controller.get_agent(agent_name, project_name)
            if agent is None:
                raise ValueError(f"Agent with name doesn't exist: {agent_name}")
            existing_chatbot_rec.agent_id = agent.id

        if model_name is not None:
            # TODO check model_name
            existing_chatbot_rec.model_name = model_name
        if database_id is not None:
            # TODO check database_id
            existing_chatbot_rec.database_id = database_id

        task = db.Tasks.query.filter(
            db.Tasks.object_type == self.OBJECT_TYPE,
            db.Tasks.object_id == existing_chatbot_rec.id,
            db.Tasks.company_id == ctx.company_id,
        ).first()

        if task is not None:
            if is_running is not None:
                task.active = is_running

            # reload task
            task.reload = True

        if params is not None:
            # Merge params on update
            existing_params = existing_chatbot_rec.params or {}
            params.update(existing_params)
            existing_chatbot_rec.params = params
        db.session.commit()

        return existing_chatbot_rec

    def delete_chatbot(self, chatbot_name: str, project_name: str = 'mindsdb'):
        '''
        Deletes a chatbot by name.

        Parameters:
            chatbot_name (str): The name of the chatbot to delete
            project_name (str): The name of the containing project
        '''

        bot = self.get_chatbot(chatbot_name, project_name)
        if bot is None:
            raise Exception(f"Chat bot doesn't exist: {chatbot_name}")

        bot_rec = db.ChatBots.query.get(bot['id'])

        task = db.Tasks.query.filter(
            db.Tasks.object_type == self.OBJECT_TYPE,
            db.Tasks.object_id == bot_rec.id,
            db.Tasks.company_id == ctx.company_id,
        ).first()

        if task is not None:
            db.session.delete(task)

        db.session.delete(bot_rec)

        db.session.commit()



from mindsdb.integrations.handlers.matrixone_handler import Handler as MatrixOneHandler


class D0ltHandler(MatrixOneHandler):
    """
    This handler handles connection and execution of the SQL  statements With D0lt.
    """

    name = "d0lt"

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



import os
import time
from typing import Any, Dict

import pandas as pd
import webzio
from dotty_dict import dotty
from mindsdb_sql import parse_sql

from mindsdb.integrations.handlers.webz_handler.webz_tables import (
    WebzPostsTable,
    WebzReviewsTable,
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.utilities import log
from mindsdb.utilities.config import Config

logger = log.getLogger(__name__)


class WebzHandler(APIHandler):
    """A class for handling connections and interactions with the Webz API."""

    API_CALL_EXEC_LIMIT_SECONDS = 60
    AVAILABLE_CONNECTION_ARGUMENTS = ["token"]

    def __init__(self, name: str = None, **kwargs):
        """Registers all tables and prepares the handler for an API connection.

        Args:
            name: (str): The handler name to use
        """
        super().__init__(name)

        args = kwargs.get("connection_data", {})
        self.connection_args = self._read_connection_args(name, **args)

        self.client = None
        self.is_connected = False
        self.max_page_size = 100

        self._register_table(WebzPostsTable.TABLE_NAME, WebzPostsTable(self))
        self._register_table(WebzReviewsTable.TABLE_NAME, WebzReviewsTable(self))

    def _read_connection_args(self, name: str = None, **kwargs) -> Dict[str, Any]:
        """Read the connection arguments by following the order of precedence below:

        1. PARAMETERS object
        2. Environment Variables
        3. MindsDB Config File

        """
        filtered_args = {}
        handler_config = Config().get(f"{name.lower()}_handler", {})
        for k in type(self).AVAILABLE_CONNECTION_ARGUMENTS:
            if k in kwargs:
                filtered_args[k] = kwargs[k]
            elif f"{name.upper()}_{k.upper()}" in os.environ:
                filtered_args[k] = os.environ[f"{name.upper()}_{k.upper()}"]
            elif k in handler_config:
                filtered_args[k] = handler_config[k]
        return filtered_args

    def connect(self) -> object:
        """Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected and self.client is not None:
            return self.client

        webzio.config(token=self.connection_args["token"])
        self.client = webzio
        self.is_connected = True
        return self.client

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        response = StatusResponse(False)
        try:
            webzio_client = self.connect()
            webzio_client.query("filterWebContent", {"q": "AI", "size": 1})
            response.success = True
        except Exception as e:
            response.error_message = f"Error connecting to Webz api: {e}."

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str = None) -> Response:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, api's json etc)
        Returns:
            HandlerResponse
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)

    def _parse_item(self, item, output_colums):
        dotted_item = dotty(item)
        return {field.replace(".", "__"): dotted_item[field] for field in output_colums}

    def call_webz_api(
        self, method_name: str = None, params: Dict = None
    ) -> pd.DataFrame:
        """Calls the API method with the given params.

        Returns results as a pandas DataFrame.

        Args:
            method_name (str): Method name to call
            params (Dict): Params to pass to the API call
        """
        table_name = method_name
        table = self._tables[table_name]

        client = self.connect()

        left = None
        count_results = None

        data = []
        limit_exec_time = time.time() + type(self).API_CALL_EXEC_LIMIT_SECONDS

        if "size" in params:
            count_results = params["size"]

        #  GET param q is mandatory, so in order to collect all data,
        # it's needed to use as a query an asterisk (*)
        if "q" not in params:
            params["q"] = "*"

        while True:
            if time.time() > limit_exec_time:
                raise RuntimeError("Handler request timeout error")

            if count_results is not None:
                left = count_results - len(data)
                if left == 0:
                    break
                elif left < 0:
                    # got more results that we need
                    data = data[:left]
                    break

                if left > self.max_page_size:
                    params["size"] = self.max_page_size
                else:
                    params["size"] = left

            logger.debug(
                f"Calling Webz API: {table.ENDPOINT} with params ({params})"
            )

            output = (
                client.query(table.ENDPOINT, params)
                if len(data) == 0
                else client.get_next()
            )
            for item in output.get(table_name, []):
                data.append(self._parse_item(item, table.OUTPUT_COLUMNS))

        df = pd.DataFrame(data)
        return df



from collections import OrderedDict
from typing import Optional

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE, HandlerResponse
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
import pandas as pd
import jaydebeapi as jd

logger = log.getLogger(__name__)

class MaxDBHandler(DatabaseHandler):
    """
       This handler handles connection and execution of the SAP MaxDB  statements.
       """

    name = 'maxdb'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """ Initialize the handler
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.kwargs = kwargs
        self.parser = parse_sql
        self.connection_config = connection_data
        self.database = connection_data['database']
        self.host = connection_data['host']
        self.port = connection_data['port']
        self.user = connection_data['user']
        self.password = connection_data['password']
        self.jdbc_location = connection_data['jdbc_location']
        self.connection = None
        self.is_connected = False

    def __del__(self):
        """
        Destructor for the SAP MaxDB class.
        """
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Establishes a connection to the SAP MaxDB server.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected:
            return self.connection

        jdbc_url = f"jdbc:sapdb://{self.host}:{self.port}/{self.database}"
        jdbc_class = 'com.sap.dbtech.jdbc.DriverSapDB'

        self.connection = jd.connect(jdbc_class, jdbc_url, [self.user, self.password], self.jdbc_location)
        self.is_connected = True
        return self.connection

    def disconnect(self):
        """ Close any existing connections
        Should switch self.is_connected.
        """
        if self.is_connected is False:
            return
        try:
            self.connection.close()
            self.is_connected = False
        except Exception as e:
            logger.error(f"Error while disconnecting to {self.database}, {e}")

        return

    def check_connection(self) -> StatusResponse:
        """ Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to database {self.database}, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> HandlerResponse:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)
        Returns:
            HandlerResponse
        """
        need_to_close = self.is_connected is False
        conn = self.connect()
        with conn.cursor() as cur:
            try:
                cur.execute(query)
                if cur.description:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                self.connection.commit()
            except Exception as e:
                logger.error(f'Error running query: {query} on {self.database}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )
                self.connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        renderer = SqlalchemyRender("postgres")
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Gets a list of table names in the database.

        Returns:
            list: A list of table names in the database.
        """

        query = f"SELECT TABLENAME FROM DOMAIN.TABLES WHERE TYPE = 'TABLE' AND SCHEMANAME = '{self.user}'"
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name: str) -> Response:
        """
        Gets a list of column names in the specified table.

        Args:
            table_name (str): The name of the table to get column names from.

        Returns:
            list: A list of column names in the specified table.
        """

        query = f"SELECT COLUMNNAME,DATATYPE FROM DOMAIN.COLUMNS WHERE TABLENAME ='{table_name}'"
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={'name': 'column_name', 'type': 'data_type'})
        return self.native_query(query)


connection_args = OrderedDict(
        host={
            'type': ARG_TYPE.STR,
            'description': 'The host name or IP address of the SAP MaxDB server.'
        },
        port={
            'type': ARG_TYPE.INT,
            'description': 'The TCP/IP port of the SAP MaxDB. Must be an integer.'
        },
        database={
            'type': ARG_TYPE.STR,
            'description': 'The database name to use when connecting with the SAP MaxDB server.'
        },
        jdbc_location={
            'type': ARG_TYPE.STR,
            'description': 'The location of the jar file which contains the JDBC class.'
        },
        user={
            'type': ARG_TYPE.STR,
            'description': 'The user name used to authenticate with the SAP MaxDB server.'
        },
        password={
            'type': ARG_TYPE.STR,
            'description': 'The password to authenticate the user with the SAP MaxDB server.'
        }
)


connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=7210,
    user='DBADMIN',
    password='password',
    database="MAXDB",
    jdbc_location='/Users/marsid/Desktop/sapdbc.jar',
)



import re
import ast as py_ast
import typing as t

from mindsdb_sql.parser.ast import OrderBy, Identifier, Star, Select, Constant, BinaryOperation, Tuple, Latest


class MongoToAst:

    """
      Converts query mongo to AST format
    """

    def from_mongoqeury(self, query):
        # IS NOT USED YET AND NOT FINISHED

        collection = query.collection

        filter, projection = None, None
        sort, limit, skip = None, None, None
        for step in query.pipeline:
            if step['method'] == 'find':
                filter = step['args'][0]
                if len(step) > 1:
                    projection = step['args'][1]
            elif step['method'] == 'sort':
                sort = step['args'][0]
            elif step['method'] == 'limit':
                limit = step['args'][0]
            elif step['method'] == 'skip':
                skip = step['args'][0]

        return self.find(collection, filter=filter,
                         sort=sort, projection=projection,
                         limit=limit, skip=skip)

    def find(self, collection: t.Union[list, str],
             filter=None, sort=None, projection=None,
             limit=None, skip=None, **kwargs):
        # https://www.mongodb.com/docs/v4.2/reference/method/db.collection.find/

        order_by = None
        if sort is not None:
            # sort is dict
            order_by = []
            for col, direction in sort.items():
                order_by.append(
                    OrderBy(
                        field=Identifier(parts=[col]),
                        direction='DESC' if direction == -1 else 'ASC'
                    )
                )

        if projection is not None:
            targets = []
            for col, alias in projection.items():
                # it is only identifiers
                if isinstance(alias, str):
                    alias = Identifier(parts=[alias])
                else:
                    alias = None
                targets.append(
                    Identifier(path_str=col, alias=alias)
                )
        else:
            targets = [Star()]

        where = None
        if filter is not None:
            where = self.convert_filter(filter)

        # convert to AST node
        #   collection can be string or list
        if isinstance(collection, list):
            collection = Identifier(parts=collection)
        else:
            collection = Identifier(path_str=collection)

        node = Select(
            targets=targets,
            from_table=collection,
            where=where,
            order_by=order_by,
        )
        if limit is not None:
            node.limit = Constant(value=limit)

        if skip is not None and skip != 0:
            node.offset = Constant(value=skip)

        return node

    def convert_filter(self, filter):
        cond_ops = {
            '$and': 'and',
            '$or': 'or',
        }

        ast_filter = None
        for k, v in filter.items():
            if k in ('$or', '$and'):
                # suppose it is one key in dict

                op = cond_ops[k]

                nodes = []
                for cond in v:
                    nodes.append(self.convert_filter(cond))

                if len(nodes) == 1:
                    return nodes[0]

                # compose as tree
                arg1 = nodes[0]
                for node in nodes[1:]:
                    arg1 = BinaryOperation(op=op, args=[arg1, node])

                return arg1
            if k in ('$where', '$expr'):
                # try to parse simple expression like 'this.saledate > this.latest'
                return MongoWhereParser(v).to_ast()

            # is filter
            arg1 = Identifier(parts=[k])

            op, value = self.handle_filter(v)
            arg2 = Constant(value=value)
            ast_com = BinaryOperation(op=op, args=[arg1, arg2])
            if ast_filter is None:
                ast_filter = ast_com
            else:
                ast_filter = BinaryOperation(op='and', args=[
                    ast_filter,
                    ast_com
                ])
        return ast_filter

    def handle_filter(self, value):
        ops = {
            '$ge': '>=',
            '$gt': '>',
            '$lt': '<',
            '$le': '<=',
            '$ne': '!=',
            '$eq': '='
        }
        in_ops = {
            '$in': 'in',
            '$nin': 'not in'
        }

        if isinstance(value, dict):
            key, value = list(value.items())[0]
            if key in ops:
                op = ops[key]
                return op, value

            if key in in_ops:
                op = in_ops[key]
                if not isinstance(value, list):
                    raise NotImplementedError(f'Unknown type {key}, {value}')
                value = Tuple(value)

                return op, value

            raise NotImplementedError(f'Unknown type {key}')

        elif isinstance(value, list):
            raise NotImplementedError(f'Unknown filter {value}')
        else:
            # is simple type
            op = '='
            value = value
            return op, value


class MongoWhereParser:
    def __init__(self, query):
        self.query = query

    def to_ast(self):
        # parse as python string
        # replace '=' with '=='
        query = re.sub(r'([^=><])=([^=])', r'\1==\2', self.query)

        tree = py_ast.parse(query, mode='eval')
        return self.process(tree.body)

    def process(self, node):

        if isinstance(node, py_ast.BoolOp):
            # is AND or OR
            op = node.op.__class__.__name__
            # values can be more than 2
            arg1 = self.process(node.values[0])
            for val1 in node.values[1:]:
                arg2 = self.process(val1)
                arg1 = BinaryOperation(op=op, args=[arg1, arg2])

            return arg1

        if isinstance(node, py_ast.Compare):
            # it is
            if len(node.ops) != 1:
                raise NotImplementedError(f'Multiple ops {node.ops}')
            op = self.compare_op(node.ops[0])
            arg1 = self.process(node.left)
            arg2 = self.process(node.comparators[0])
            return BinaryOperation(op=op, args=[arg1, arg2])

        if isinstance(node, py_ast.Name):
            # is special operator: latest, ...
            if node.id == 'latest':
                return Latest()

        if isinstance(node, py_ast.Constant):
            # it is constant
            return Constant(value=node.value)

        # ---- python 3.7 objects -----
        if isinstance(node, py_ast.Str):
            return Constant(value=node.s)

        if isinstance(node, py_ast.Num):
            return Constant(value=node.n)

        # -----------------------------

        if isinstance(node, py_ast.Attribute):
            # is 'this.field' - is attribute
            if node.value.id != 'this':
                raise NotImplementedError(f'Unknown variable {node.value.id}')
            return Identifier(parts=[node.attr])

        raise NotImplementedError(f'Unknown node {node}')

    def compare_op(self, op):

        opname = op.__class__.__name__

        # TODO: in, not

        ops = {
            'Eq': '=',
            'NotEq': '!=',
            'Gt': '>',
            'Lt': '<',
            'GtE': '>=',
            'LtE': '<=',
        }
        if opname not in ops:
            raise NotImplementedError(f'Unknown $where op: {opname}')
        return ops[opname]

    @staticmethod
    def test(cls):
        assert cls('this.a ==1 and "te" >= latest').to_string() == "a = 1 AND 'te' >= LATEST"



import sys
from dataclasses import dataclass
from autosklearn.metrics import *
from autosklearn.metrics import Scorer


@dataclass(frozen=True)
class BaseConfig:
    time_left_for_this_task: int = 3600
    per_run_time_limit: int = None
    n_jobs: int = None
    metric: Scorer = 'accuracy'
    ensemble_size: int = None
    ensemble_nbest: int = 50
    initial_configurations_via_metalearning: int = 25
    resampling_strategy: str = 'holdout'

    def __post_init__(self):
        object.__setattr__(self, 'metric', getattr(sys.modules[__name__], self.metric))


@dataclass(frozen=True)
class ClassificationConfig(BaseConfig):
    pass


@dataclass(frozen=True)
class RegressionConfig(BaseConfig):
    pass



import pandas as pd
import json
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.handlers.vertex_handler.vertex_client import VertexClient
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class VertexHandler(BaseMLEngine):
    """Handler for the Vertex Google AI cloud API"""

    name = "Vertex"

    def create(self, target, args=None, **kwargs):
        """Logs in to Vertex and deploy a pre-trained model to an endpoint.

        If the endpoint already exists for the model, we do nothing.

        If the endpoint does not exist, we create it and deploy the model to it.
        The runtime for this is long, it took 15 minutes for a small model.
        """
        assert "using" in args, "Must provide USING arguments for this handler"
        args = args["using"]

        model_name = args.pop("model_name")
        custom_model = args.pop("custom_model", False)

        service_account_info = self.engine_storage.json_get('service_account')

        # get vertex args from handler then update args from model
        vertex_args = self.engine_storage.json_get('args')
        vertex_args.update(args)

        vertex = VertexClient(service_account_info, vertex_args)

        model = vertex.get_model_by_display_name(model_name)
        if not model:
            raise Exception(f"Vertex model {model_name} not found")
        endpoint_name = model_name + "_endpoint"
        if vertex.get_endpoint_by_display_name(endpoint_name):
            logger.info(f"Endpoint {endpoint_name} already exists, skipping deployment")
        else:
            logger.info(f"Starting deployment at {endpoint_name}")
            endpoint = vertex.deploy_model(model)
            endpoint.display_name = endpoint_name
            endpoint.update()
            logger.info(f"Endpoint {endpoint_name} deployed")

        predict_args = {}
        predict_args["target"] = target
        predict_args["endpoint_name"] = endpoint_name
        predict_args["custom_model"] = custom_model
        self.model_storage.json_set("predict_args", predict_args)
        self.model_storage.json_set("vertex_args", vertex_args)

    def predict(self, df, args=None):
        """Predict using the deployed model by calling the endpoint."""

        if "__mindsdb_row_id" in df.columns:
            df.drop("__mindsdb_row_id", axis=1, inplace=True)  # TODO is this required?

        predict_args = self.model_storage.json_get("predict_args")
        vertex_args = self.model_storage.json_get("vertex_args")
        service_account_info = self.engine_storage.json_get('service_account')

        vertex = VertexClient(service_account_info, vertex_args)
        results = vertex.predict_from_df(predict_args["endpoint_name"], df, custom_model=predict_args["custom_model"])

        if predict_args["custom_model"]:
            return pd.DataFrame(results.predictions, columns=[predict_args["target"]])
        else:
            return pd.DataFrame(results.predictions)

    def create_engine(self, connection_args):
        if 'service_account' not in connection_args:
            raise KeyError('"service_account" parameter is required')
        service_account = connection_args.pop('service_account')
        if isinstance(service_account, str):
            # convert to json
            service_account = json.loads(service_account)
        else:
            # unescape new lines in private_key
            service_account['private_key'] = service_account['private_key'].replace('\\n', '\n')

        self.engine_storage.json_set('args', connection_args)
        self.engine_storage.json_set('service_account', service_account)



from http import HTTPStatus

from flask import request
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.interfaces.skills.skills_controller import SkillsController


def create_skill(project_name, skill):
    skills_controller = SkillsController()
    for required_field in ['name', 'type', 'params']:
        if required_field not in skill:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing field',
                'Missing "{}" field for skill'.format(required_field)
            )
    name = skill['name']
    type = skill['type']
    params = skill['params']

    try:
        existing_skill = skills_controller.get_skill(name, project_name)
    except ValueError:
        # Project needs to exist
        return http_error(
            HTTPStatus.NOT_FOUND,
            'Project not found',
            f'Project with name {project_name} does not exist'
        )

    if existing_skill is not None:
        return http_error(
            HTTPStatus.CONFLICT,
            'Skill already exists',
            f'Skill with name {name} already exists. Please use a different name'
        )

    new_skill = skills_controller.add_skill(name, project_name, type, params)
    return new_skill.as_dict(), HTTPStatus.CREATED


@ns_conf.route('/<project_name>/skills')
class SkillsResource(Resource):
    @ns_conf.doc('list_skills')
    def get(self, project_name):
        ''' List all skills'''
        skills_controller = SkillsController()
        try:
            all_skills = skills_controller.get_skills(project_name)
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )
        return [skill.as_dict() for skill in all_skills]

    @ns_conf.doc('create_skill')
    def post(self, project_name):
        '''Create a skill'''

        # Check required request format.
        if 'skill' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "skill" parameter in POST body'
            )
        skill = request.json['skill']
        return create_skill(project_name, skill)


@ns_conf.route('/<project_name>/skills/<skill_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('skill_name', 'Name of the skill')
class SkillResource(Resource):
    @ns_conf.doc('get_skill')
    def get(self, project_name, skill_name):
        '''Gets a skill by name'''
        skills_controller = SkillsController()
        try:
            existing_skill = skills_controller.get_skill(skill_name, project_name)
        except ValueError:
            # Project needs to exist
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        if existing_skill is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Skill not found',
                f'Skill with name {skill_name} not found.'
            )
        return existing_skill.as_dict()

    @ns_conf.doc('update_skill')
    def put(self, project_name, skill_name):
        '''Updates a skill by name, creating one if it doesn't exist'''
        skills_controller = SkillsController()

        # Check required request format.
        if 'skill' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "skill" parameter in POST body'
            )

        try:
            existing_skill = skills_controller.get_skill(skill_name, project_name)
        except ValueError:
            # Project needs to exist
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        skill = request.json['skill']
        if existing_skill is None:
            # Use same name as provided in URL.
            skill['name'] = skill_name
            return create_skill(project_name, skill)

        new_name = skill.get('name', None)
        new_type = skill.get('type', None)
        new_params = skill.get('params', None)
        updated_skill = skills_controller.update_skill(
            skill_name,
            new_name=new_name,
            project_name=project_name,
            type=new_type,
            params=new_params)
        return updated_skill.as_dict()

    @ns_conf.doc('delete_skill')
    def delete(self, project_name, skill_name):
        '''Deletes a skill by name'''
        skills_controller = SkillsController()
        try:
            existing_skill = skills_controller.get_skill(skill_name, project_name)
        except ValueError:
            # Project needs to exist
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        if existing_skill is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Skill not found',
                f'Skill with name {skill_name} not found.'
            )
        skills_controller.delete_skill(skill_name, project_name)
        return '', HTTPStatus.NO_CONTENT



from http import HTTPStatus

from flask_restx import Resource, abort
from sqlalchemy.exc import NoResultFound


from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController


@ns_conf.route('/')
class ProjectsList(Resource):
    @ns_conf.doc('list_projects')
    def get(self):
        ''' List all projects '''
        session = SessionController()

        projects = [
            {'name': i}
            for i in session.datahub.get_projects_names()
        ]
        return projects


@ns_conf.route('/<project_name>')
class ProjectsGet(Resource):
    @ns_conf.doc('get_project')
    def get(self, project_name):
        '''Gets a project by name'''
        session = SessionController()

        try:
            project = session.database_controller.get_project(project_name)
        except NoResultFound:
            abort(HTTPStatus.NOT_FOUND, f'Project name {project_name} does not exist')

        return {
            'name': project.name
        }



import os

from mindsdb.api.mysql.mysql_proxy.executor.executor import Executor
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class ExecutorClientFactory:
    def __init__(self):
        self.host = os.environ.get("MINDSDB_EXECUTOR_SERVICE_HOST")
        self.port = os.environ.get("MINDSDB_EXECUTOR_SERVICE_PORT")
        if self.host is not None and self.port is not None:
            try:
                from mindsdb.api.mysql.mysql_proxy.executor.executor_grpc_client import ExecutorClientGRPC
                self.client_class = ExecutorClientGRPC
            except (ImportError, ModuleNotFoundError):
                logger.error("to use microservice mode please install 'pip install mindsdb[grpc]'")
                self.host = None
                self.port = None

    def __call__(self, session, sqlserver):
        if self.host is None or self.port is None:
            logger.info(
                "%s.__call__: no post/port to ExecutorService have provided. Local Executor instance will use",
                self.__class__.__name__,
            )
            return Executor(session, sqlserver)

        logger.info("%s.__call__: api to communicate with db services - gRPC",
                    self.__class__.__name__,
                    )

        return self.client_class(session, sqlserver)


ExecutorClient = ExecutorClientFactory()



from functools import lru_cache
from typing import List, Union

import pandas as pd
import torch
from pydantic import BaseModel
from sentence_transformers import SentenceTransformer


class Parameters(BaseModel):
    embeddings_model_name: str


@lru_cache()
def load_embeddings_model(embeddings_model_name):
    try:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        embedding_model = SentenceTransformer(
            model_name_or_path=embeddings_model_name, device=device
        )
    except ValueError:
        raise ValueError(
            f"The {embeddings_model_name}  is not supported, please select a valid option from Hugging Face Hub!"
        )
    return embedding_model


class DfLoader:
    def __init__(self, data_frame: pd.DataFrame, page_content_column: str):
        self._data_frame = data_frame
        self._page_content_column = page_content_column

    def load(self) -> List[str]:
        """takes a Dataframe column and loads it as a list of texts"""
        texts = []
        for n_row, frame in self._data_frame[self._page_content_column].iteritems():
            if pd.notnull(frame):
                # ignore rows with None values
                texts.append(frame)
        return texts


def df_to_documents(
    df: pd.DataFrame, page_content_columns: Union[List[str], str]
) -> List[str]:
    """Given a subset of columns, converts a dataframe into a list of Documents"""
    texts = []

    if isinstance(page_content_columns, str):
        page_content_columns = [page_content_columns]

    for _, page_content_column in enumerate(page_content_columns):
        if page_content_column not in df.columns.tolist():
            raise ValueError(
                f"page_content_column {page_content_column} not in dataframe columns"
            )

        loader = DfLoader(data_frame=df, page_content_column=page_content_column)
        texts.extend(loader.load())

    return texts



from typing import Optional
from collections import OrderedDict

import pandas as pd
import pyodbc

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from sqlalchemy_access.base import AccessDialect
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

logger = log.getLogger(__name__)


class AccessHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Microsoft Access statements.
    """

    name = 'access'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'access'
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.connection = pyodbc.connect(
            r'Driver={Microsoft Access Driver (*.mdb, *.accdb)};DBQ=' + self.connection_data['db_file']
        )
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to SQLite {self.connection_data["db_file"]}, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cursor:
            try:
                cursor.execute(query)
                result = cursor.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame.from_records(
                            result,
                            columns=[x[0] for x in cursor.description]
                        )
                    )

                else:
                    response = Response(RESPONSE_TYPE.OK)
                    connection.commit()
            except Exception as e:
                logger.error(f'Error running query: {query} on {self.connection_data["db_file"]}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        renderer = SqlalchemyRender(AccessDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        connection = self.connect()
        with connection.cursor() as cursor:
            df = pd.DataFrame([table.table_name for table in cursor.tables(tableType='Table')], columns=['table_name'])

        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )

        return response

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        connection = self.connect()
        with connection.cursor() as cursor:
            df = pd.DataFrame(
                [(column.column_name, column.type_name) for column in cursor.columns(table=table_name)],
                columns=['column_name', 'data_type']
            )

        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )

        return response


connection_args = OrderedDict(
    db_file={
        'type': ARG_TYPE.STR,
        'description': 'The database file where the data will be stored.'
    }
)

connection_args_example = OrderedDict(
    db_file='C:\\Users\\minurap\\Documents\\example_db.accdb'
)



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .eventstoredb_handler import EventStoreDB as Handler
from .__about__ import __version__ as version

title = 'EventStoreDB'
name = 'eventstoredb'
type = HANDLER_TYPE.DATA
icon_path = 'icon.svg'

__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'icon_path'
]



import unittest
from mindsdb.integrations.handlers.impala_handler.impala_handler import ImpalaHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class ImpalaHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            'user':'<UID>',
            'password':'<P455w0rd>',
            'host':'127.0.0.1',
            'port':21050,
            'database':'temp'
    
        }
        cls.handler = ImpalaHandler('test_impala_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_connect(self):
        assert self.handler.connect()
    
    def test_2_create_table(self):
        query = "CREATE Table Car(Name Varchar, Price Integer);"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR 

    def test_3_insert(self):
        query = "INSERT INTO Car ('Tata SUV', 860000)"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR 

    def test_4_native_query_select(self):
        query = "SELECT * FROM Car;"
        result = self.handler.query(query)
        assert result.type is RESPONSE_TYPE.TABLE 

    def test_5_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is  RESPONSE_TYPE.TABLE

    def test_6_get_columns(self):
        columns = self.handler.get_columns('Car')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



import urllib
import tempfile
from pathlib import Path

import requests


def get_column_in_case(columns, name):
    '''
    '''
    candidates = []
    name_lower = name.lower()
    for column in columns:
        if column.lower() == name_lower:
            candidates.append(column)
    if len(candidates) != 1:
        return None
    return candidates[0]


def download_file(url):
    try:
        parse_result = urllib.parse.urlparse(url)
        scheme = parse_result.scheme
    except ValueError:
        raise Exception(f'Invalid url: {url}')
    except Exception as e:
        raise Exception(f'URL parsing error: {e}')
    temp_dir = tempfile.mkdtemp(prefix='mindsdb_file_download_')
    if scheme == '':
        raise Exception(f"Unknown url schema: {url}")

    response = requests.get(url)
    temp_file_path = Path(temp_dir).joinpath('file')
    with open(str(temp_file_path), 'wb')as file:
        file.write(response.content)
    return str(temp_file_path)



import os

import mindsdb.interfaces.storage.db as db
from mindsdb.api.mysql.mysql_proxy.executor.executor_grpc_wrapper import (
    ExecutorServiceServicer,
)
from mindsdb.utilities.config import Config
from mindsdb.utilities import log

if __name__ == "__main__":
    config = Config()
    db.init()
    logger = log.getLogger(__name__)
    app = ExecutorServiceServicer()
    port = int(os.environ.get("PORT", 5500))
    host = os.environ.get("HOST", "0.0.0.0")
    logger.info("Running Executor service: host=%s, port=%s", host, port)
    app.run(debug=True, host=host, port=port)



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""
import math
import struct

from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import (
    DEFAULT_CAPABILITIES,
    NULL_VALUE,
    ONE_BYTE_ENC,
    THREE_BYTE_ENC,
    TWO_BYTE_ENC,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class Datum:
    def __init__(self, type, value=None):
        # TODO other types: float, timestamp
        self.type = type
        self.value = b""
        self.var_type = self.type.split("<")[0]
        self.var_len = self.type.split("<")[1].replace(">", "")

        if value is not None:
            self.set(value)

    def set(self, value):
        self.value = value

    def setFromBuff(self, buff):
        start = 0

        if self.var_len == "lenenc":
            start = 1
            ln_enc = buff[0]
            if int(ln_enc) <= ONE_BYTE_ENC[0]:
                start = 0
                end = 1
            elif int(ln_enc) == TWO_BYTE_ENC[0]:
                end = 3
            elif int(ln_enc) == THREE_BYTE_ENC[0]:
                end = 4
            elif ln_enc:
                end = 9

            num_str = buff[start:end]
            if end > 9:
                logger.error("Cant decode integer greater than 8 bytes")
                return buff[end - 1 :]  # noqa: E203

            for j in range(8 - (end - start)):
                num_str += b"\0"

            if self.var_type == "int":
                self.value = struct.unpack("i", num_str)
                return buff[end:]

            if self.var_type in ["byte", "string"]:
                length = struct.unpack("Q", num_str)[0]
                self.value = buff[end : (length + end)]  # noqa: E203
                return buff[(length + end) :]  # noqa: E203

        if self.var_len == "EOF":
            length = len(buff)
            self.var_len = str(length)
            self.value = buff
            return ""
        else:
            length = self.var_len

        if self.type == "string<NUL>":
            for j, x in enumerate(buff):
                if int(x) == 0:
                    length = j + 1
                    break

        length = int(length)
        if self.var_type in ["byte", "string"]:
            end = length
            self.value = buff[:end]
        else:  # if its an integer
            end = length
            num_str = buff[:end]
            if end > 8:
                logger.error("cant decode integer greater than 8 bytes")
                return buff[end:]
            for j in range(8 - end):
                num_str += b"\0"
            self.value = struct.unpack("Q", num_str)[0]
        if str(self.var_len) == "NUL":
            self.value = self.value[:-1]
        return buff[end:]

    def lenencInt(self, value):
        byte_count = int(math.ceil(math.log((value + 1), 2) / 8))
        if byte_count == 0:
            return b"\0"
        if value < NULL_VALUE[0]:
            return struct.pack("i", value)[:1]
        if value >= NULL_VALUE[0] and byte_count <= 2:
            return TWO_BYTE_ENC + struct.pack("i", value)[:2]
        if byte_count <= 3:
            return THREE_BYTE_ENC + struct.pack("i", value)[:3]
        if byte_count <= 8:
            return THREE_BYTE_ENC + struct.pack("Q", value)[:8]

    def toStringPacket(self):
        if self.type == "string<packet>":
            return self.value.get_packet_string()

        if self.type in ["string<EOF>", "byte<EOF>"]:
            length = int(len(self.value))
            self.var_len = length
            if length == 0:
                return b""
            else:
                return struct.pack(
                    "{len}s".format(len=self.var_len), bytes(self.value, "utf-8")
                )[:length]

        if self.type == "string<NUL>":
            return bytes(self.value, "utf-8") + struct.pack("b", 0)

        if self.var_len.isdigit():
            length = int(self.var_len)

            if self.var_type == "int":
                # little endian format
                return struct.pack("Q", self.value)[:length]
            if self.var_type == "string":
                return struct.pack(self.var_len + "s", bytes(self.value, "utf-8"))[
                    :length
                ]
            if self.var_type == "byte":
                return struct.pack(self.var_len + "s", self.value)[:length]

        elif self.var_len == "lenenc":
            if self.value is None:
                return NULL_VALUE
            if self.var_type == "int":
                return self.lenencInt(self.value)

            if self.var_type in ["byte", "string"]:
                value = (
                    self.value.decode() if isinstance(self.value, bytes) else self.value
                )
                if isinstance(value, str) is False and value is not None:
                    value = str(value)

                val_len = len(value.encode("utf8"))

                byte_count = int(math.ceil(math.log((val_len + 1), 2) / 8))
                if val_len < NULL_VALUE[0]:
                    return self.lenencInt(val_len) + bytes(value, "utf-8")
                if val_len >= NULL_VALUE[0] and byte_count <= 2:
                    return (
                        TWO_BYTE_ENC
                        + struct.pack("i", val_len)[:2]
                        + bytes(value, "utf-8")
                    )
                if byte_count <= 3:
                    return (
                        THREE_BYTE_ENC
                        + struct.pack("i", val_len)[:3]
                        + bytes(value, "utf-8")
                    )
                if byte_count <= 8:
                    return (
                        THREE_BYTE_ENC
                        + struct.pack("Q", val_len)[:8]
                        + bytes(value, "utf-8")
                    )


def test():
    import pprint

    u = Datum("int<8>", DEFAULT_CAPABILITIES >> 16)
    pprint.pprint(u.toStringPacket())


# only run the test if this file is called from debugger
if __name__ == "__main__":
    test()



from typing import List

import pandas as pd
from mindsdb_sql.parser import ast

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class WebzBaseAPITable(APITable):

    ENDPOINT = None
    OUTPUT_COLUMNS = []
    SORTABLE_COLUMNS = []
    TABLE_NAME = None

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Selects data from the API and returns it as a pandas DataFrame

        Returns dataframe representing the API results.

        Args:
            query (ast.Select): SQL SELECT query

        """
        conditions = extract_comparison_conditions(query.where)
        params = {}

        for op, arg1, arg2 in conditions:
            if op != "=":
                raise NotImplementedError(f"Unsupported Operator: {op}")
            elif arg1 == "query":
                params["q"] = arg2
            else:
                raise NotImplementedError(f"Unknown clause: {arg1}")

        if query.order_by:
            if len(query.order_by) > 1:
                raise ValueError("Unsupported to order by multiple fields")
            order_item = query.order_by[0]
            sort_column = ".".join(order_item.field.parts[1:])
            # make sure that column is sortable
            if sort_column not in type(self).SORTABLE_COLUMNS:
                raise ValueError(f"Order by unknown column {sort_column}")
            params.update({"sort": sort_column, "order": order_item.direction.lower()})

        if query.limit is not None:
            params["size"] = query.limit.value
        result = self.handler.call_webz_api(
            method_name=type(self).TABLE_NAME, params=params
        )

        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError(f"Unknown query target {type(target)}")

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            return pd.DataFrame([], columns=columns)

        # add absent columns
        for col in set(columns) & set(result.columns) ^ set(columns):
            result[col] = None

        # filter by columns
        result = result[columns]

        # Rename columns
        for target in query.targets:
            if target.alias:
                result.rename(
                    columns={target.parts[-1]: str(target.alias)}, inplace=True
                )
        return result

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
            List of columns

        """
        return [column.replace(".", "__") for column in type(self).OUTPUT_COLUMNS]


class WebzPostsTable(WebzBaseAPITable):
    """To interact with structured posts data from news articles, blog posts and online discussions
    provided through the Webz.IO API.

    """

    ENDPOINT = "filterWebContent"
    OUTPUT_COLUMNS = [
        "thread.uuid",
        "thread.url",
        "thread.site_full",
        "thread.site",
        "thread.site_section",
        "thread.section_title",
        "thread.title",
        "thread.title_full",
        "thread.published",
        "thread.replies_count",
        "thread.participants_count",
        "thread.site_type",
        "thread.main_image",
        "thread.country",
        "thread.site_categories",
        "thread.social.facebook.likes",
        "thread.social.facebook.shares",
        "thread.social.facebook.comments",
        "thread.social.gplus.shares",
        "thread.social.pinterest.shares",
        "thread.social.linkedin.shares",
        "thread.social.stumbledupon.shares",
        "thread.social.vk.shares",
        "thread.performance_score",
        "thread.domain_rank",
        "thread.domain_rank_updated",
        "thread.reach.per_million",
        "thread.reach.page_views",
        "thread.reach.updated",
        "uuid",
        "url",
        "ord_in_thread",
        "parent_url",
        "author",
        "published",
        "title",
        "text",
        "language",
        "external_links",
        "external_images",
        "rating",
        "entities.persons",
        "entities.organizations",
        "entities.locations",
        "crawled",
    ]
    SORTABLE_COLUMNS = [
        "crawled",
        "relevancy",
        "social.facebook.likes",
        "social.facebook.shares",
        "social.facebook.comments",
        "social.gplus.shares",
        "social.pinterest.shares",
        "social.linkedin.shares",
        "social.stumbledupon.shares",
        "social.vk.shares",
        "replies_count",
        "participants_count",
        "performance_score",
        "published",
        "thread.published",
        "domain_rank",
        "ord_in_thread",
        "rating",
    ]
    TABLE_NAME = "posts"


class WebzReviewsTable(WebzBaseAPITable):
    """To interact with structured reviews data from hundreds of review sites,
    provided through the Webz.IO API.

    """

    ENDPOINT = "reviewFilter"
    OUTPUT_COLUMNS = [
        "item.uuid",
        "item.url",
        "item.site_full",
        "item.site",
        "item.site_section",
        "item.section_title",
        "item.title",
        "item.title_full",
        "item.published",
        "item.reviews_count",
        "item.reviewers_count",
        "item.main_image",
        "item.country",
        "item.site_categories",
        "item.domain_rank",
        "item.domain_rank_updated",
        "uuid",
        "url",
        "ord_in_thread",
        "author",
        "published",
        "title",
        "text",
        "language",
        "external_links",
        "rating",
        "crawled",
    ]
    SORTABLE_COLUMNS = [
        "crawled",
        "relevancy",
        "reviews_count",
        "reviewers_count",
        "spam_score",
        "domain_rank",
        "ord_in_thread",
        "rating",
    ]
    TABLE_NAME = "reviews"



from mindsdb.utilities import log

from mindsdb.integrations.libs.api_handler import APIHandler

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb_sql import parse_sql

from mindsdb.integrations.handlers.email_handler.email_tables import EmailsTable
from mindsdb.integrations.handlers.email_handler.email_helpers import EmailClient

logger = log.getLogger(__name__)

class EmailHandler(APIHandler):
    """A class for handling connections and interactions with Email (send and search).

    Attributes:
        "email": "Email address used to login to the SMTP and IMAP servers.",
        "password": "Password used to login to the SMTP and IMAP servers.",
        "smtp_server": "SMTP server to be used for sending emails. Default value is 'smtp.gmail.com'.",
        "smtp_port": "Port number for the SMTP server. Default value is 587.",
        "imap_server": "IMAP server to be used for reading emails. Default value is 'imap.gmail.com'."
    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        emails = EmailsTable(self)
        self._register_table('emails', emails)

    def connect(self):
        """Authenticate with the email servers using credentials."""

        if self.is_connected is True:
            return self.connection

        try:
            self.connection = EmailClient(**self.connection_data)
        except Exception as e:
            logger.error(f'Error connecting to email api: {e}!')
            raise e

        self.is_connected = True
        return self.connection

    def check_connection(self) -> StatusResponse:

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True

        except Exception as e:
            response.error_message = f'Error connecting to Email: {e}. '
            logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)




from mindsdb.integrations.handlers.snowflake_handler.snowflake_handler import SnowflakeHandler
import unittest
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb_sql import parse_sql


class SnowflakeHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "us-east-1.snowflakecomputing.com",
            "port": "443",
            "user": "",
            "password": "",
            "database": "DEMO_DB",
            "warehouse": "COMPUTE_WH",
            "account": "us-east-1.aws",
            "schema": "PUBLIC",
            "protocol": "https"
        }
        cls.handler = SnowflakeHandler('test_snowflake_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_get_tables(self):
        tbls = self.handler.get_tables()
        assert tbls.type is not RESPONSE_TYPE.ERROR

    def test_2_get_columns(self):
        tbls = self.handler.get_columns('home_rentals')
        assert tbls.type is not RESPONSE_TYPE.ERROR

    def test_3_select_native(self):
        query = "SELECT * FROM home_rentals WHERE sqft=484"
        ast = parse_sql(query)
        result = self.handler.query(ast)
        assert result.type is RESPONSE_TYPE.TABLE



import os
import json
import requests
import datetime as dt
from flask import request
from shutil import copyfile

from mindsdb.utilities import log

from .exceptions import AuthException

from google_auth_oauthlib.flow import Flow

from google.oauth2.credentials import Credentials
from google.auth.transport.requests import Request

logger = log.getLogger(__name__)


class GoogleOAuth2Manager:
    def __init__(self, handler_stroage: str, scopes: list, credentials_file: str = None, credentials_url: str = None, code: str = None):
        self.handler_storage = handler_stroage
        self.scopes = scopes
        self.credentials_file = credentials_file
        self.credentials_url = credentials_url
        self.code = code

    def get_oauth2_credentials(self):
        creds = None

        if self.credentials_file or self.credentials_url:
            # get the current directory and checks tokens & creds
            curr_dir = self.handler_storage.folder_get('config')

            creds_file = os.path.join(curr_dir, 'creds.json')
            secret_file = os.path.join(curr_dir, 'secret.json')

            if os.path.isfile(creds_file):
                creds = Credentials.from_authorized_user_file(creds_file, self.scopes)

            if not creds or not creds.valid:
                logger.debug("Credentials do not exist or are invalid, attempting to authorize again")

                if self._download_secret_file(secret_file):
                    # save to storage
                    self.handler_storage.folder_sync('config')
                else:
                    raise ValueError('No valid Gmail Credentials filepath or S3 url found.')

                if creds and creds.expired and creds.refresh_token:
                    creds.refresh(Request())
                    logger.debug("Credentials refreshed successfully")
                else:
                    creds = self._execute_google_auth_flow(secret_file, self.scopes, self.code)
                    logger.debug("New credentials obtained")

                self._save_credentials_to_file(creds, creds_file)
                logger.debug(f"saved session credentials to {creds_file}")
                self.handler_storage.folder_sync('config')

        return creds

    def _download_secret_file(self, secret_file):
        # if credentials_url is set, attempt to download the file
        # this will be given preference over credentials_file
        if self.credentials_url:
            response = requests.get(self.credentials_url)
            if response.status_code == 200:
                with open(secret_file, 'w') as creds:
                    creds.write(response.text)
                return True
            else:
                logger.error("Failed to get credentials from S3", response.status_code)

        # if credentials_file is set, attempt to copy the file
        if self.credentials_file and os.path.isfile(self.credentials_file):
            copyfile(self.credentials_file, secret_file)
            return True
        return False
    
    def _execute_google_auth_flow(self, secret_file, scopes, code=None):
        flow = Flow.from_client_secrets_file(secret_file, scopes)

        flow.redirect_uri = request.headers['ORIGIN'] + '/verify-auth'

        if code:
            flow.fetch_token(code=code)
            creds = flow.credentials
            return creds
        else:
            auth_url = flow.authorization_url()[0]
            raise AuthException(f'Authorisation required. Please follow the url: {auth_url}', auth_url=auth_url)
    
    def _save_credentials_to_file(self, creds, file_path):
        with open(file_path, 'w') as token:
            data = self._convert_credentials_to_dict(creds)
            token.write(json.dumps(data))

    def _convert_credentials_to_dict(self, credentials):
        return {
            'token': credentials.token,
            'refresh_token': credentials.refresh_token,
            'token_uri': credentials.token_uri,
            'client_id': credentials.client_id,
            'client_secret': credentials.client_secret,
            'scopes': credentials.scopes,
            'expiry': dt.datetime.strftime(credentials.expiry, '%Y-%m-%dT%H:%M:%S')
        }


from collections import OrderedDict

import duckdb
import pandas as pd
from duckdb import DuckDBPyConnection
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.const import (
    HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE,
)
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class DuckDBHandler(DatabaseHandler):
    """This handler handles connection and execution of the DuckDB statements."""

    name = 'duckdb'

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'postgresql'
        self.connection_data = kwargs.get('connection_data')
        self.renderer = SqlalchemyRender('postgres')

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> DuckDBPyConnection:
        """Connect to a DuckDB database.

        Returns:
            DuckDBPyConnection: The database connection.
        """

        if self.is_connected is True:
            return self.connection

        args = {
            'database': self.connection_data.get('database'),
            'read_only': self.connection_data.get('read_only'),
        }

        self.connection = duckdb.connect(**args)
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """Close the database connection."""

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """Check the connection to the DuckDB database.

        Returns:
            StatusResponse: Connection success status and error message if an error occurs.
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(
                f'Error connecting to DuckDB {self.connection_data["database"]}, {e}!'
            )
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """Execute a SQL query.

        Args:
            query (str): The SQL query to execute.

        Returns:
            Response: The query result.
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        cursor = connection.cursor()

        try:
            cursor.execute(query)

            result = cursor.fetchall()
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result, columns=[x[0] for x in cursor.description]
                    ),
                )
            else:
                connection.commit()
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            logger.error(
                f'Error running query: {query} on {self.connection_data["database"]}!'
            )
            response = Response(RESPONSE_TYPE.ERROR, error_message=str(e))

        cursor.close()
        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """Render and execute a SQL query.

        Args:
            query (ASTNode): The SQL query.

        Returns:
            Response: The query result.
        """

        query_str = self.renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """Get a list of all the tables in the database.

        Returns:
            Response: Names of the tables in the database.
        """

        q = 'SHOW TABLES;'
        result = self.native_query(q)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name: str) -> Response:
        """Get details about a table.

        Args:
            table_name (str): Name of the table to retrieve details of.

        Returns:
            Response: Details of the table.
        """

        query = f'DESCRIBE {table_name};'
        return self.native_query(query)


connection_args = OrderedDict(
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database file to read and write from. The special value :memory: (default) can be used to create an in-memory database.',
    },
    read_only={
        'type': ARG_TYPE.BOOL,
        'description': 'A flag that specifies if the connection should be made in read-only mode.',
    },
)

connection_args_example = OrderedDict(database='db.duckdb', read_only=True)



import tiktoken

from typing import Callable
from mindsdb_sql import parse_sql, Insert

from langchain.utilities import GoogleSerperAPIWrapper
from langchain.prompts import PromptTemplate
from langchain.agents import load_tools, Tool

from langchain.chains.llm import LLMChain
from langchain.text_splitter import CharacterTextSplitter
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.chains import ReduceDocumentsChain, MapReduceDocumentsChain

from mindsdb.interfaces.skills.skill_tool import make_tools_from_skill


# Individual tools
# Note: all tools are defined in a closure to pass required args (apart from LLM input) through it, as custom tools don't allow custom field assignment.  # noqa
def get_exec_call_tool(llm, executor, model_kwargs) -> Callable:
    def mdb_exec_call_tool(query: str) -> str:
        try:
            ast_query = parse_sql(query.strip('`'), dialect='mindsdb')
            ret = executor.execute_command(ast_query)
            if ret.data is None and ret.error_code is None:
                return ''
            data = ret.data  # list of lists
            data = '\n'.join([  # rows
                '\t'.join(      # columns
                    str(row) if isinstance(row, str) else [str(value) for value in row]
                ) for row in data
            ])
        except Exception as e:
            data = f"mindsdb tool failed with error:\n{str(e)}"   # let the agent know

        # summarize output if needed
        data = summarize_if_overflowed(data, llm, model_kwargs['max_tokens'])

        return data
    return mdb_exec_call_tool

def get_exec_metadata_tool(llm, executor, model_kwargs) -> Callable:
    def mdb_exec_metadata_call(query: str) -> str:
        try:
            parts = query.replace('`', '').split('.')
            assert 1 <= len(parts) <= 2, 'query must be in the format: `integration` or `integration.table`'

            integration = parts[0]
            integrations = executor.session.integration_controller
            handler = integrations.get_handler(integration)

            if len(parts) == 1:
                df = handler.get_tables().data_frame
                data = f'The integration `{integration}` has {df.shape[0]} tables: {", ".join(list(df["TABLE_NAME"].values))}'  # noqa

            if len(parts) == 2:
                df = handler.get_tables().data_frame
                table_name = parts[-1]
                try:
                    table_name_col = 'TABLE_NAME' if 'TABLE_NAME' in df.columns else 'table_name'
                    mdata = df[df[table_name_col] == table_name].iloc[0].to_list()
                    if len(mdata) == 3:
                        _, nrows, table_type = mdata
                        data = f'Metadata for table {table_name}:\n\tRow count: {nrows}\n\tType: {table_type}\n'
                    elif len(mdata) == 2:
                        nrows = mdata
                        data = f'Metadata for table {table_name}:\n\tRow count: {nrows}\n'
                    else:
                        data = f'Metadata for table {table_name}:\n'
                    fields = handler.get_columns(table_name).data_frame['Field'].to_list()
                    types = handler.get_columns(table_name).data_frame['Type'].to_list()
                    data += f'List of columns and types:\n'
                    data += '\n'.join([f'\tColumn: `{field}`\tType: `{typ}`' for field, typ in zip(fields, types)])
                except:
                    data = f'Table {table_name} not found.'
        except Exception as e:
            data = f"mindsdb tool failed with error:\n{str(e)}"  # let the agent know

        # summarize output if needed
        data = summarize_if_overflowed(data, llm, model_kwargs['max_tokens'])

        return data
    return mdb_exec_metadata_call

def get_mdb_write_tool(executor) -> Callable:
    def mdb_write_call(query: str) -> str:
        try:
            query = query.strip('`')
            ast_query = parse_sql(query.strip('`'), dialect='mindsdb')
            if isinstance(ast_query, Insert):
                _ = executor.execute_command(ast_query)
                return "mindsdb write tool executed successfully"
        except Exception as e:
            return f"mindsdb write tool failed with error:\n{str(e)}"
    return mdb_write_call

def _setup_standard_tools(tools, llm, executor, model_kwargs):
    all_standard_tools = []
    langchain_tools = []
    for tool in tools:
        if tool == 'mindsdb_read':
            mdb_tool = Tool(
                name="MindsDB",
                func=get_exec_call_tool(llm, executor, model_kwargs),
                description="useful to read from databases or tables connected to the mindsdb machine learning package. the action must be a valid simple SQL query, always ending with a semicolon. For example, you can do `show databases;` to list the available data sources, and `show tables;` to list the available tables within each data source."  # noqa
            )

            mdb_meta_tool = Tool(
                name="MDB-Metadata",
                func=get_exec_metadata_tool(llm, executor, model_kwargs),
                description="useful to get column names from a mindsdb table or metadata from a mindsdb data source. the command should be either 1) a data source name, to list all available tables that it exposes, or 2) a string with the format `data_source_name.table_name` (for example, `files.my_table`), to get the table name, table type, column names, data types per column, and amount of rows of the specified table."  # noqa
            )
            all_standard_tools.append(mdb_tool)
            all_standard_tools.append(mdb_meta_tool)
        elif tool == 'mindsdb_write':
            mdb_write_tool = Tool(
                name="MDB-Write",
                func=get_mdb_write_tool(executor),
                description="useful to write into data sources connected to mindsdb. command must be a valid SQL query with syntax: `INSERT INTO data_source_name.table_name (column_name_1, column_name_2, [...]) VALUES (column_1_value_row_1, column_2_value_row_1, [...]), (column_1_value_row_2, column_2_value_row_2, [...]), [...];`. note the command always ends with a semicolon. order of column names and values for each row must be a perfect match. If write fails, try casting value with a function, passing the value without quotes, or truncating string as needed.`."  # noqa
            )
            all_standard_tools.append(mdb_write_tool)
        else:
            langchain_tools.append(tool)
    
    all_standard_tools += load_tools(langchain_tools)
    return all_standard_tools

# Collector
def setup_tools(llm, model_kwargs, pred_args, executor, default_agent_tools, openai_api_key):
    toolkit = pred_args['tools'] if pred_args['tools'] is not None else default_agent_tools

    standard_tools = []
    function_tools = []

    for tool in toolkit:
        if isinstance(tool, str):
            standard_tools.append(tool)
        else:
            # user defined custom functions
            function_tools.append(tool)

    skill_tools = []
    skills = pred_args.get('skills', [])
    for skill in skills:
        skill_tools += make_tools_from_skill(skill, llm, openai_api_key, executor)

    tools = _setup_standard_tools(standard_tools, llm, executor, model_kwargs)
    if model_kwargs.get('serper_api_key', False):
        search = GoogleSerperAPIWrapper(serper_api_key=model_kwargs.pop('serper_api_key'))
        tools.append(Tool(
            name="Intermediate Answer (serper.dev)",
            func=search.run,
            description="useful for when you need to search the internet (note: in general, use this as a last resort)"  # noqa
        ))

    for tool in function_tools:
        tools.append(Tool(
            name=tool['name'],
            func=tool['func'],
            description=tool['description'],
        ))
    tools = tools + skill_tools

    return tools


# Helpers
def summarize_if_overflowed(data, llm, max_tokens, budget_multiplier=0.8) -> str:
    """
        This helper retries with a summarized version of the
        output if the previous call fails due to the token limit being exceeded.

        We trigger summarization when the token count exceeds the limit times a multiplier to be conservative.
    """
    # tokenize data for length check
    # note: this is a rough estimate, as the tokenizer used in each LLM may be different
    encoding = tiktoken.get_encoding("gpt2")
    n_tokens = len(encoding.encode(data))

    # map-reduce given token budget
    if n_tokens > max_tokens * budget_multiplier:
        # map
        map_template = """The following is a set of documents
                {docs}
                Based on this list of docs, please identify the main themes
                Helpful Answer:"""
        map_prompt = PromptTemplate.from_template(map_template)
        map_chain = LLMChain(llm=llm, prompt=map_prompt)

        # reduce
        reduce_template = """The following is set of summaries:
                {doc_summaries}
                Take these and distill it into a final, consolidated summary of the main themes.
                Helpful Answer:"""
        reduce_prompt = PromptTemplate.from_template(reduce_template)
        reduce_chain = LLMChain(llm=llm, prompt=reduce_prompt)
        combine_documents_chain = StuffDocumentsChain(
            llm_chain=reduce_chain, document_variable_name="doc_summaries"
        )
        reduce_documents_chain = ReduceDocumentsChain(
            combine_documents_chain=combine_documents_chain,
            collapse_documents_chain=combine_documents_chain,
            token_max=max_tokens * budget_multiplier,  # applies for each group of documents
        )
        map_reduce_chain = MapReduceDocumentsChain(
            llm_chain=map_chain,
            reduce_documents_chain=reduce_documents_chain,
            document_variable_name="docs",
            return_intermediate_steps=False,
        )
        # split
        text_splitter = CharacterTextSplitter.from_tiktoken_encoder(chunk_size=1000, chunk_overlap=0)
        docs = text_splitter.create_documents([data])
        split_docs = text_splitter.split_documents(docs)

        # run chain
        data = map_reduce_chain.run(split_docs)
    return data



from typing import Optional, Dict

import pandas as pd
from mindsdb.utilities import log

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.utilities.handler_utils import get_api_key

from mindsdb.integrations.handlers.twelve_labs_handler.settings import TwelveLabsHandlerConfig
from mindsdb.integrations.handlers.twelve_labs_handler.twelve_labs_api_client import TwelveLabsAPIClient


logger = log.getLogger(__name__)

# TODO: move to config
BASE_URL = "https://api.twelvelabs.io/v1.1"
DEFAULT_ENGINE = "marengo2.5"
DEFAULT_WAIT_DURATION = 5


class TwelveLabsHandler(BaseMLEngine):
    """
    Integration with the Twelve Labs API.
    """

    name = 'twelve_labs'

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        """
        Validates the create arguments.
        Args:
            target (str): name of the target column
            args (dict): dictionary of create arguments
            **kwargs: arbitrary keyword arguments.
        Returns:
            None
        """
        # check for USING clause
        if 'using' not in args:
            # TODO: update Exception to InsufficientParametersException
            raise Exception("Twelve Labs engine requires a USING clause! Refer to its documentation for more details.")
        else:
            # get USING args
            args = args['using']
            TwelveLabsHandlerConfig(**args)

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        # get USING args and add target
        args = args['using']
        args['target'] = target

        # get api key
        api_key = get_api_key(
            api_name=self.name,
            create_args=args,
            engine_storage=self.engine_storage,
        )

        # initialize TwelveLabsAPIClient
        twelve_labs_api_client = TwelveLabsAPIClient(api_key=api_key)

        # update args with api key
        args['api_key'] = api_key

        # get index if it exists
        index_id = twelve_labs_api_client.get_index_by_name(index_name=args['index_name'])

        # create index if it doesn't exist
        if not index_id:
            logger.info(f"Index {args['index_name']} does not exist. Creating index.")
            index_id = twelve_labs_api_client.create_index(
                index_name=args['index_name'],
                engine_id=args['engine_id'] if 'engine_id' in args else None,
                index_options=args['index_options'],
                addons=args['addons'] if 'addons' in args else []
            )

        else:
            logger.info(f"Index {args['index_name']} already exists. Using existing index.")

        # store index_id in args
        args['index_id'] = index_id

        # create video indexing tasks for all video files or video urls
        # video urls will be given precedence
        # check if video_urls_col has been set and use it to get the video urls
        if 'video_urls_col' in args:
            logger.info("video_urls_col has been set, therefore, it will be given precedence.")
            video_urls = df[args['video_urls_col']].tolist()

        # else, check if video_files_col has been set and use it to get the video files
        elif 'video_files_col' in args:
            logger.info("video_urls_col has not been set, therefore, video_files_col will be used.")
            video_files = df[args['video_files_col']].tolist()

        # else, check if video_urls or video_files have been set and use them
        else:
            logger.info("video_urls_col and video_files_col have not been set, therefore, video_urls and video_files will be used.")
            video_urls = args['video_urls'] if 'video_urls' in args else None
            video_files = args['video_files'] if 'video_files' in args else None

        # if video_urls and video_files are not set, then raise an exception
        if not video_urls and not video_files:
            logger.error("Neither video_urls_col, video_files_col, video_urls nor video_files have been set.")
            raise RuntimeError("Neither video_urls_col, video_files_col, video_urls nor video_files have been set. Please set one of them.")

        task_ids = twelve_labs_api_client.create_video_indexing_tasks(
            index_id=index_id,
            video_urls=video_urls,
            video_files=video_files,
        )

        # poll for video indexing tasks to complete
        twelve_labs_api_client.poll_for_video_indexing_tasks(task_ids=task_ids)

        # store args in model_storage
        self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        # get args from model_storage
        args = self.model_storage.json_get('args')

        # get api key
        api_key = get_api_key(
            api_name=self.name,
            create_args=args,
            engine_storage=self.engine_storage,
        )

        # initialize TwelveLabsAPIClient
        twelve_labs_api_client = TwelveLabsAPIClient(api_key=api_key)

        # get search query
        query = df['query'].tolist()[0]

        # check if task is search
        if args['task'] == 'search':
            # search for query in index
            data = twelve_labs_api_client.search_index(
                index_id=args['index_id'],
                query=query,
                search_options=args['search_options']
            )

            # TODO: pick only the necessary columns?
            # TODO: structure nested columns?
            # metadata = ['score', 'start', 'end', 'video_id', 'confidence']
            # df_metadata = pd.json_normalize(data, record_path='metadata', meta=metadata, record_prefix='metadata_')
            # df_modules = pd.json_normalize(data, record_path='modules', meta=metadata, record_prefix='modules_')
            # df_predictions = pd.merge(df_metadata, df_modules, on=metadata)
            # return df_predictions
            return pd.json_normalize(data).add_prefix(args['target'] + '_')

        else:
            raise NotImplementedError(f"Task {args['task']} is not supported.")



OPENAI_API_BASE = 'https://api.openai.com/v1'

CHAT_MODELS = (
    'gpt-3.5-turbo',
    'gpt-3.5-turbo-16k',
    'gpt-3.5-turbo-instruct',
    'gpt-4',
    'gpt-4-32k',
    'gpt-4-1106-preview',
)
COMPLETION_MODELS = ('babbage-002', 'davinci-002')
FINETUNING_MODELS = ('gpt-3.5-turbo', 'babbage-002', 'davinci-002')
COMPLETION_LEGACY_BASE_MODELS = ('davinci', 'curie', 'babbage', 'ada')

FINETUNING_LEGACY_MODELS = COMPLETION_LEGACY_BASE_MODELS
COMPLETION_LEGACY_MODELS = (
    COMPLETION_LEGACY_BASE_MODELS
    + tuple(f'text-{model}-001' for model in COMPLETION_LEGACY_BASE_MODELS)
    + ('text-davinci-002', 'text-davinci-003')
)

EMBEDDING_MODELS = (
    ('text-embedding-ada-002',)
    + tuple(f'text-similarity-{model}-001' for model in COMPLETION_LEGACY_BASE_MODELS)
    + tuple(f'text-search-{model}-query-001' for model in COMPLETION_LEGACY_BASE_MODELS)
    + tuple(f'text-search-{model}-doc-001' for model in COMPLETION_LEGACY_BASE_MODELS)
    + tuple(f'code-search-{model}-text-001' for model in COMPLETION_LEGACY_BASE_MODELS)
    + tuple(f'code-search-{model}-code-001' for model in COMPLETION_LEGACY_BASE_MODELS)
)

IMAGE_MODELS = ('dall-e-2', 'dall-e-3')



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'hostInfo': helpers.is_true}

    # NOTE answer is not full, but looks like this info is enough
    result = {
        'system': {
            'cpuAddrSize': 64,
            'memSizeMB': 15951,
            'numCores': 4,
            'cpuArch': "x86_64",
            'numaEnabled': False
        },
        'os': {
            'type': "Linux",
            'name': "Ubuntu",
            'version': "20.04"
        },
        'extra': {
        },
        'ok': 1
    }


responder = Responce()



import unittest
from mindsdb.integrations.handlers.google_books_handler.google_books_handler import GoogleBooksHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class GoogleBooksHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "credentials": "C:\\Users\\panagiotis\\Desktop\\GitHub\\mindsdb\\mindsdb\\integrations\\handlers"
                               "\\google_books_handler\\credentials.json"
            }
        }
        cls.handler = GoogleBooksHandler('test_google_books_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_2_select_volume_query(self):
        query = "SELECT summary FROM my_books.volumes WHERE q = 'Harry Potter'"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_3_select_bookshelves_query(self):
        query = "SELECT title FROM my_books.bookshelves WHERE shelf > 1"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_4_get_columns(self):
        columns = self.handler.get_columns('id')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



from collections import OrderedDict
from typing import Any, List, Optional
from itertools import zip_longest

from qdrant_client import QdrantClient, models
import pandas as pd

from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    FilterOperator,
    TableField,
    VectorStoreHandler,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class QdrantHandler(VectorStoreHandler):
    """Handles connection and execution of the Qdrant statements."""

    name = "qdrant"

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        connection_data = kwargs.get("connection_data").copy()
        # Qdrant offers several configuration and optmization options at the time of collection creation
        # Since the create table statement doesn't have a way to pass these options
        # We are requiring the user to pass these options in the connection_data
        # These options are documented here. https://qdrant.github.io/qdrant/redoc/index.html#tag/collections/operation/create_collection
        self.collection_config = connection_data.pop("collection_config")
        self.connect(**connection_data)

    def connect(self, **kwargs):
        """Connect to a Qdrant instance.
        A Qdrant client can be instantiated with a REST, GRPC interface or in-memory for testing.
        To use the in-memory instance, specify the location argument as ':memory:'."""
        if self.is_connected:
            return self._client

        try:
            self._client = QdrantClient(**kwargs)
            self.is_connected = True
            return self._client
        except Exception as e:
            logger.error(f"Error instantiating a Qdrant client: {e}")
            self.is_connected = False

    def disconnect(self):
        """Close the database connection."""
        if self.is_connected:
            self._client.close()
            self._client = None
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """Check the connection to the Qdrant database.

        Returns:
            StatusResponse: Indicates if the connection is alive
        """
        need_to_close = not self.is_connected

        try:
            # Using a trivial operation to get the connection status
            # As there isn't a universal ping method for the REST, GRPC and in-memory interface
            self._client.get_locks()
            response_code = StatusResponse(True)
        except Exception as e:
            logger.error(f"Error connecting to a Qdrant instance: {e}")
            response_code = StatusResponse(False, error_message=str(e))
        finally:
            if response_code.success and need_to_close:
                self.disconnect()
            if not response_code.success and self.is_connected:
                self.is_connected = False

        return response_code

    def drop_table(self, table_name: str, if_exists=True) -> HandlerResponse:
        """Delete a collection from the Qdrant Instance.

        Args:
            table_name (str): The name of the collection to be dropped
            if_exists (bool, optional): Throws an error if this value is set to false and the collection doesn't exist. Defaults to True.

        Returns:
            HandlerResponse: _description_
        """
        result = self._client.delete_collection(table_name)
        if result or if_exists:
            return Response(resp_type=RESPONSE_TYPE.OK)
        else:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Table {table_name} does not exist!",
            )

    def get_tables(self) -> HandlerResponse:
        """Get the list of collections in the Qdrant instance.

        Returns:
            HandlerResponse: The common query handler response with a list of table names
        """
        collection_response = self._client.get_collections()
        collections_name = pd.DataFrame(
            columns=["table_name"],
            data=[collection.name for collection in collection_response.collections],
        )
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=collections_name)

    def get_columns(self, table_name: str) -> HandlerResponse:
        try:
            _ = self._client.get_collection(table_name)
        except ValueError:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Table {table_name} does not exist!",
            )
        return super().get_columns(table_name)

    def insert(
        self, table_name: str, data: pd.DataFrame, columns: List[str] = None
    ) -> HandlerResponse:
        """Handler for the insert query

        Args:
            table_name (str): The name of the table to be inserted into
            data (pd.DataFrame): The data to be inserted
            columns (List[str], optional): Columns to be inserted into. Unused as the values are derived from the "data" argument. Defaults to None.

        Returns:
            HandlerResponse: The common query handler response
        """
        assert len(data[TableField.ID.value]) == len(data[TableField.EMBEDDINGS.value]), "Number of ids and embeddings must be equal"

        # Qdrant doesn't have a distinction between documents and metadata
        # Any data that is to be stored should be placed in the "payload" field
        data = data.to_dict(orient="list")
        payloads = []
        content_list = data[TableField.CONTENT.value]
        metadata_list = data[TableField.METADATA.value]

        for document, metadata in zip_longest(content_list, metadata_list, fillvalue=None):
            payload = {}

            # Insert the document with a "document" key in the payload
            if document is not None:
                payload["document"] = document

            # Unpack all the metadata fields into the payload
            if metadata is not None:
                payload = {**payload, **metadata}

            if payload:
                payloads.append(payload)

        # IDs can be either integers or strings(UUIDs)
        # The following step ensures proper type of numberic values
        ids = [int(id) if str(id).isdigit() else id for id in data[TableField.ID.value]]
        self._client.upsert(table_name, points=models.Batch(
            ids=ids,
            vectors=data[TableField.EMBEDDINGS.value],
            payloads=payloads
        ))

        return Response(resp_type=RESPONSE_TYPE.OK)

    def create_table(self, table_name: str, if_not_exists=True) -> HandlerResponse:
        """Create a collection with the given name in the Qdrant database.

        Args:
            table_name (str): Name of the table(Collection) to be created
            if_not_exists (bool, optional): Throws an error if this value is set to false and the collection already exists. Defaults to True.

        Returns:
            HandlerResponse: The common query handler response
        """
        try:
            # Create a collection with the collection name and collection_config set during __init__
            self._client.create_collection(table_name, self.collection_config)
        except ValueError:
            if if_not_exists is False:
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message=f"Table {table_name} already exists!",
                )

        return Response(resp_type=RESPONSE_TYPE.OK)

    def _get_qdrant_filter(self, operator: FilterOperator, value: Any) -> dict:
        """ Map the filter operator to the Qdrant filter
            We use a match and not a dict so as to conditionally construct values
            With a dict, all the values the values will be constructed
            Generating models.Range() with a str type value fails

        Args:
            operator (FilterOperator): FilterOperator specified in the query. Eg >=, <=, =
            value (Any): Value specified in the query

        Raises:
            Exception: If an unsupported operator is specified

        Returns:
            dict: A dict of Qdrant filtering clauses
        """
        if operator == FilterOperator.EQUAL:
            return {"match": models.MatchValue(value=value)}
        elif operator == FilterOperator.NOT_EQUAL:
            return {"match": models.MatchExcept(**{"except": [value]})}
        elif operator == FilterOperator.LESS_THAN:
            return {"range": models.Range(lt=value)}
        elif operator == FilterOperator.LESS_THAN_OR_EQUAL:
            return {"range": models.Range(lte=value)}
        elif operator == FilterOperator.GREATER_THAN:
            return {"range": models.Range(gt=value)}
        elif operator == FilterOperator.GREATER_THAN_OR_EQUAL:
            return {"range": models.Range(gte=value)}
        else:
            raise Exception(f"Operator {operator} is not supported by Qdrant!")

    def _translate_filter_conditions(
        self, conditions: List[FilterCondition]
    ) -> Optional[dict]:
        """
        Translate a list of FilterCondition objects a dict that can be used by Qdrant.
        Filtering clause docs can be found here: https://qdrant.tech/documentation/concepts/filtering/
        E.g.,
        [
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.LESS_THAN,
                value=7132423,
            ),
            FilterCondition(
                column="metadata.created_at",
                op=FilterOperator.GREATER_THAN,
                value=2323432,
            )
        ]
        -->
        models.Filter(
        must=[
            models.FieldCondition(
                key="created_at",
                match=models.Range(lt=7132423),
            ),
            models.FieldCondition(
                key="created_at",
                match=models.Range(gt=2323432),
            ),
          ]
        )
        """
        # We ignore all non-metadata conditions
        if conditions is None:
            return None
        filter_conditions = [
            condition
            for condition in conditions
            if condition.column.startswith(TableField.METADATA.value)
        ]
        if len(filter_conditions) == 0:
            return None

        qdrant_filters = []
        for condition in filter_conditions:
            payload_key = condition.column.split(".")[-1]
            qdrant_filters.append(
                models.FieldCondition(key=payload_key, **self._get_qdrant_filter(condition.op, condition.value))
            )

        return models.Filter(must=qdrant_filters) if qdrant_filters else None

    def update(
        self, table_name: str, data: pd.DataFrame, columns: List[str] = None
    ) -> HandlerResponse:
        """
        Update data in the Qdrant database.
        TODO: Update for vector DBs has not been implemented.
        Ref: https://github.com/mindsdb/mindsdb/blob/a870ba93b0afee234e48c0268489a94a6e6fd5f7/mindsdb/integrations/libs/vectordatabase_handler.py#L273-L277
        """
        return super().update(table_name, data, columns)

    def select(self, table_name: str, columns: Optional[List[str]] = None, conditions: Optional[List[FilterCondition]] = None, offset: int = 0, limit: int = 10,) -> HandlerResponse:
        """Select query handler
           Eg: SELECT * FROM qdrant.test_table

        Args:
            table_name (str): The name of the table to be queried
            columns (Optional[List[str]], optional): List of column names specified in the query. Defaults to None.
            conditions (Optional[List[FilterCondition]], optional): List of "where" conditionals. Defaults to None.
            offset (int, optional): Offset the results by the provided value. Defaults to 0.
            limit (int, optional): Number of results to return. Defaults to 10.

        Returns:
            HandlerResponse: The common query handler response
        """

        # Validate and set offset and limit as None is passed if not set in the query
        offset = offset if offset is not None else 0
        limit = limit if limit is not None else 10

        # Full scroll if no where conditions are specified
        if not conditions:
            results = self._client.scroll(table_name, limit=limit, offset=offset)
            payload = self._process_select_results(results[0], columns)
            return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=payload)

        # Filter conditions
        vector_filter = [condition.value for condition in conditions if condition.column == TableField.SEARCH_VECTOR.value]
        id_filters = [condition.value for condition in conditions if condition.column == TableField.ID.value]
        query_filters = self._translate_filter_conditions(conditions)

        # Prefer returning results by IDs first
        if id_filters:
            results = self._client.retrieve(table_name, ids=id_filters)
        # Followed by the search_vector value
        elif vector_filter:
            # Perform a similarity search with the first vector filter
            results = self._client.search(table_name, query_vector=vector_filter[0], query_filter=query_filters or None, limit=limit, offset=offset)
        elif query_filters:
            results = self._client.scroll(table_name, scroll_filter=query_filters, limit=limit, offset=offset)[0]

        # Process results
        payload = self._process_select_results(results, columns)
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=payload)

    def _process_select_results(self, results=None, columns=None):
        """Private method to process the results of a select query

        Args:
            results: A List[Records] or List[ScoredPoint]. Defaults to None
            columns: List of column names specified in the query. Defaults to None

        Returns:
            Dataframe: A processed pandas dataframe
        """
        ids, documents, metadata, distances = [], [], [], []

        for result in results:
            ids.append(result.id)
            # The documents and metadata are stored as a dict in the payload
            documents.append(result.payload["document"])
            metadata.append({k: v for k, v in result.payload.items() if k != "document"})

            # Score is only available for similarity search results
            if "score" in result:
                distances.append(result.score)

        payload = {
            TableField.ID.value: ids,
            TableField.CONTENT.value: documents,
            TableField.METADATA.value: metadata,
        }

        # Filter result columns
        if columns:
            payload = {
                column: payload[column]
                for column in columns
                if column != TableField.EMBEDDINGS.value and column in payload
            }

        # If the distance list is empty, don't add it to the result
        if distances:
            payload[TableField.DISTANCE.value] = distances

        return pd.DataFrame(payload)

    def delete(
        self, table_name: str, conditions: List[FilterCondition] = None
    ) -> HandlerResponse:
        """Delete query handler

        Args:
            table_name (str): List of column names specified in the query. Defaults to None.
            conditions (List[FilterCondition], optional): List of "where" conditionals. Defaults to None.

        Raises:
            Exception: If no conditions are specified

        Returns:
            HandlerResponse: The common query handler response
        """
        filters = self._translate_filter_conditions(conditions)
        # Get id filters
        ids = [
            condition.value
            for condition in conditions
            if condition.column == TableField.ID.value
        ] or None

        if filters is None and ids is None:
            raise Exception("Delete query must have at least one condition!")

        if ids:
            self._client.delete(table_name, points_selector=models.PointIdsList(points=ids))

        if filters:
            self._client.delete(table_name, points_selector=models.FilterSelector(filter=filters))
        return Response(resp_type=RESPONSE_TYPE.OK)


connection_args = OrderedDict(
    location={
        "type": ARG_TYPE.STR,
        "description": "If `:memory:` - use in-memory Qdrant instance. If a remote URL - connect to a remote Qdrant instance. Example: `http://localhost:6333`",
        "required": False,
    },
    url={
        "type": ARG_TYPE.STR,
        "description": "URL of Qdrant service. Either host or a string of type [scheme]<host><[port][prefix]. Ex: http://localhost:6333/service/v1",
    },
    host={
        "type": ARG_TYPE.STR,
        "description": "Host name of Qdrant service. The port and host are used to construct the connection URL.",
        "required": False,
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "Port of the REST API interface. Default: 6333",
        "required": False,
    },
    grpc_port={
        "type": ARG_TYPE.INT,
        "description": "Port of the gRPC interface. Default: 6334",
        "required": False,
    },
    prefer_grpc={
        "type": ARG_TYPE.BOOL,
        "description": "If `true` - use gPRC interface whenever possible in custom methods. Default: false",
        "required": False,
    },
    https={
        "type": ARG_TYPE.BOOL,
        "description": "If `true` - use https protocol.",
        "required": False,
    },
    api_key={
        "type": ARG_TYPE.STR,
        "description": "API key for authentication in Qdrant Cloud.",
        "required": False,
    },
    prefix={
        "type": ARG_TYPE.STR,
        "description": "If set, the value is added to the REST URL path. Example: `service/v1` will result in `http://localhost:6333/service/v1/{qdrant-endpoint}` for REST API",
        "required": False,
    },
    timeout={
        "type": ARG_TYPE.INT,
        "description": "Timeout for REST and gRPC API requests. Defaults to 5.0 seconds for REST and unlimited for gRPC",
        "required": False,
    },
    path={
        "type": ARG_TYPE.STR,
        "description": "Persistence path for a local Qdrant instance.",
        "required": False,
    },
    collection_config={
        "type": ARG_TYPE.DICT,
        "description": "Collection creation configuration. See https://qdrant.github.io/qdrant/redoc/index.html#tag/collections/operation/create_collection",
        "required": True,
    },
)

connection_args_example = {
    "location": ":memory:",
    "collection_config": {
        "size": 386,
        "distance": "Cosine"
    }
}
